#include <PamLevel2.h>
//--------------------------------------
//
//
//--------------------------------------
/**
 * Default constructor 
 */
PamTrack::PamTrack(){
    trk_track  = 0;
    calo_track = 0;
    tof_track  = 0;
    candeleteobj = 0;
};
//--------------------------------------
//
//
//--------------------------------------
/**
 * Constructor 
 */
PamTrack::PamTrack(TrkTrack* t, CaloTrkVar* c, ToFTrkVar* o){ 

    trk_track  = 0;
    calo_track = 0;
    tof_track  = 0;
//     if(t)trk_track  = new TrkTrack(*t);
//     if(c)calo_track = new CaloTrkVar(*c);
//     if(o)tof_track  = new ToFTrkVar(*o);
    if(t)trk_track  = t;
    if(c)calo_track = c;
    if(o)tof_track  = o;
    candeleteobj = 0;

};
PamTrack::PamTrack(const PamTrack& track){

    TrkTrack   *t = track.trk_track;
    CaloTrkVar *c = track.calo_track;
    ToFTrkVar  *o = track.tof_track;

    trk_track  = 0;
    calo_track = 0;
    tof_track  = 0;
    if(t)trk_track  = new TrkTrack(*t);
    if(c)calo_track = new CaloTrkVar(*c);
    if(o)tof_track  = new ToFTrkVar(*o);
    candeleteobj = 1;
    
}
void PamTrack::Clear(){

//    cout << "PamTrack::Clear() "<<candeleteobj<<endl;
    if(candeleteobj){
	if(trk_track)  trk_track->TrkTrack::Clear();
	if(calo_track) calo_track->CaloTrkVar::Clear();//???
	if(tof_track)  tof_track->ToFTrkVar::Clear();//???
    }else{
	trk_track  = 0;
	calo_track = 0;
	tof_track  = 0;	
    }
}
void PamTrack::Delete(){
//    cout << "PamTrack::Delete() "<<candeleteobj<<endl;
    if(candeleteobj){
	if(trk_track)  {
	    trk_track->TrkTrack::Clear();
	    delete trk_track;
	}
	if(calo_track){
	    calo_track->CaloTrkVar::Clear();//???
	    delete calo_track;
	}
	if(tof_track){
	    tof_track->ToFTrkVar::Clear();//???
	    delete tof_track;
	}
    }else{
	Clear();
    }
}
//--------------------------------------
//
//
//--------------------------------------
/**
 * Default Constructor 
 */
PamLevel2::PamLevel2(){
  Initialize();
};

/**
 * Constructor with given dir, list and detectors
 */
PamLevel2::PamLevel2(TString ddir,TString list,TString detlist){
  Initialize();
  GetPamTree(GetListOfLevel2Files(ddir,list),detlist);
  GetRunTree(GetListOfLevel2Files(ddir,list));
};

/**
 * Constructor with given dir and list 
 */
PamLevel2::PamLevel2(TString ddir,TString list){
  Initialize();
  GetPamTree(GetListOfLevel2Files(ddir,list),"");
  GetRunTree(GetListOfLevel2Files(ddir,list));
};


void PamLevel2::Initialize(){
	
    h0_obj    = 0;
    trk0_obj  = 0;
    calo0_obj  = 0;

    trk2_obj  = 0;
    trk1_obj  = 0;
    trkh_obj  = 0;
    calo1_obj = 0;
    calo2_obj = 0;
    tof_obj   = 0;
    trig_obj  = 0;
    s4_obj    = 0;
    nd_obj    = 0;
    ac_obj    = 0;
    orb_obj   = 0;

    run_obj   = 0;//new GL_RUN();
    soft_obj   = 0;// Emiliano
    irun = -1LL;
    runfirstentry = 0LL;
    runlastentry = 0LL;

    l0_file = NULL;
    l0_tree = NULL;
    iroot   = -1;
    dbc     = 0;

    prevshift = 0;
    
    run_tree = NULL;
    run_tree_clone = NULL;
    sel_tree = NULL;
    sel_tree_clone = NULL;
    
    irunentry = -1LL;
    pam_tree = NULL;
    for(Int_t i=0; i<NCLONES; i++ )pam_tree_clone[i]=NULL;

    host = "mysql://localhost/pamelaprod";
    user = "anonymous";
    psw = "";
    const char *pamdbhost=gSystem->Getenv("PAM_DBHOST");
    const char *pamdbuser=gSystem->Getenv("PAM_DBUSER");
    const char *pamdbpsw=gSystem->Getenv("PAM_DBPSW");
    if ( !pamdbhost ) pamdbhost = "";
    if ( !pamdbuser ) pamdbuser = "";
    if ( !pamdbpsw ) pamdbpsw = "";
    if ( strcmp(pamdbhost,"") ) host = pamdbhost;
    if ( strcmp(pamdbuser,"") ) user = pamdbuser;
    if ( strcmp(pamdbpsw,"") ) psw = pamdbpsw;

    

//    sorted_tracks = 0;//new TRefArray();
    
    CAL0 = false;
    CAL1 = true;
    CAL2 = true;
    TRK2 = true;
    TRK1 = false;
    TRKh = false;
    TRKh = false;
    TRG  = true;
    TOF  = true;
    TOF0 = false;
    S4   = true;
    ND   = true;
    AC   = true;
    ORB  = true;
    
    RUN  = true;

    SELLI = -1;

    ISGP = false;

    tsorted=0;
    timage=0;
    
    howtosort = "+CAL+TOF";
    //howtosort = "+TOF";
    sortthr = 100.;

};
/**
 * Delete the event (act as Dtor)
 */
void PamLevel2::Delete(){
	
    if(run_obj)delete run_obj;
    if(soft_obj)delete soft_obj; //Emiliano

//    cout << "void PamLevel2::Clear()"<<endl;
    if(h0_obj)   delete h0_obj;	
    if(trk0_obj) delete trk0_obj;	
    if(calo0_obj) delete calo0_obj;	
    if(trk1_obj) delete trk1_obj;	
    if(trk2_obj) delete trk2_obj;
    if(trkh_obj) delete trkh_obj;
    if(calo1_obj)delete calo1_obj;
    if(calo2_obj)delete calo2_obj;
    if(tof_obj)  delete tof_obj;
    if(trig_obj) delete trig_obj;
    if(s4_obj)   delete s4_obj;
    if(nd_obj)   delete nd_obj;
    if(ac_obj)   delete ac_obj;
    if(orb_obj)  delete orb_obj;
    
    if(tsorted){
	tsorted->Delete();
	delete tsorted;
    }
    if(timage){
	timage->Delete();
	delete timage;
    }

    if(dbc){ 
	dbc->Close();
	delete dbc;
    }
    
    if(l0_file)l0_file->Close();
    //    if(pam_tree)pam_tree->Delete();;

  if ( pam_tree ){
    //
    // we have first to find which chains we have to delete, then delete the main chain and only after delete the friend chains. Any other order causes a segfault...
    //
    TList *temp = pam_tree->GetListOfFriends();
    TList *contents  = new TList; // create chain friend list
    contents->SetOwner();
    TIter next(temp);
    TChain *questo = 0;
    while ( (questo = (TChain*) next()) ){
      TString name =  questo->GetName();
      contents->Add((TChain*)gROOT->FindObject(name.Data()));// add object to the list
    };
    //
    // deleting the main chain
    //
    pam_tree->Delete();
    //
    // deleting the friends...
    //
    TIter next2(contents);
    TChain *questa = 0;
    while ( questa = (TChain*)next2() ){
      TString name =  questa->GetName();
      questa->Delete();      
      questa=NULL;
    };
    //
  };
  pam_tree = NULL;

    if(run_tree)run_tree->Delete();;
    if(sel_tree)sel_tree->Delete();;
    for(Int_t i=0; i<NCLONES; i++ )if(pam_tree_clone[i])pam_tree_clone[i]->Delete();;
    if(run_tree_clone)run_tree_clone->Delete();;
    if(sel_tree_clone)sel_tree_clone->Delete();;
    

};

/**
 * Clear the event (NB! does not deallocate objects)
 */
void PamLevel2::Clear(){
	
//    cout << "void PamLevel2::Clear()"<<endl;

//
// This method is called once for every entry but RunInfo and SoftInfo do not change until the next run so we cannot clear them here unless we don't
// want to load them for each event even if they are the same...
//
//    if(run_obj)delete run_obj;
//    if(run_obj) run_obj->Clear();  // Emiliano: Do not deallocate run_obj here, it will give segmentation fault! call clear instead 
//    if(soft_obj) soft_obj->Clear();

    if(h0_obj)   h0_obj->Clear();	
//    if(trk0_obj) trk0_obj->Clear();	
    if(trk1_obj) trk1_obj->Clear();	
    if(trk2_obj) trk2_obj->Clear(); 
    if(trkh_obj) trkh_obj->Clear();
    if(calo0_obj) calo0_obj->Clear();	
    if(calo1_obj)calo1_obj->Clear();
    if(calo2_obj)calo2_obj->Clear();
    if(tof_obj)  tof_obj->Clear();
    if(trig_obj) trig_obj->Clear();
    if(s4_obj)   s4_obj->Clear();
    if(nd_obj)   nd_obj->Clear();
    if(ac_obj)   ac_obj->Clear();
    if(orb_obj)  orb_obj->Clear();
    
//    if(sorted_tracks)sorted_tracks->Clear(); 
//    sorted_tracks.Clear(); 

    if(tsorted){
	tsorted->Delete();
    }
    if(timage){
	timage->Delete();
    }
};

void PamLevel2::Reset(){
  //
  // First of all clear everything
  //
  Clear();
  //
  // close and reset chains and pointers
  //    
  if ( pam_tree ){
    //
    // we have first to find which chains we have to delete, then delete the main chain and only after delete the friend chains. Any other order causes a segfault...
    //
    TList *temp = pam_tree->GetListOfFriends();
    TList *contents  = new TList; // create chain friend list
    contents->SetOwner();
    TIter next(temp);
    TChain *questo = 0;
    while ( (questo = (TChain*) next()) ){
      TString name =  questo->GetName();
      contents->Add((TChain*)gROOT->FindObject(name.Data()));// add object to the list
    };
    //
    // deleting the main chain
    //
    pam_tree->Delete();
    //
    // deleting the friends...
    //
    TIter next2(contents);
    TChain *questa = 0;
    while ( questa = (TChain*)next2() ){
      TString name =  questa->GetName();
      questa->Delete();      
      questa=NULL;
    };
    //
  };
  pam_tree = NULL;
  //
  if(run_tree)run_tree->Delete();;
  run_tree = NULL;
  if(sel_tree)sel_tree->Delete();;
  sel_tree = NULL;
  //
  // Close file
  //
  if(l0_file)l0_file->Close("R");
  l0_file = NULL;
  //
  h0_obj    = 0;
  trk0_obj  = 0;
  calo0_obj  = 0;
  //
  trk2_obj  = 0;
  trk1_obj  = 0;
  trkh_obj  = 0;
  calo1_obj = 0;
  calo2_obj = 0;
  tof_obj   = 0;
  trig_obj  = 0;
  s4_obj    = 0;
  nd_obj    = 0;
  ac_obj    = 0;
  orb_obj   = 0;
  //
  // Reset run pointers
  //
  run_obj   = 0;//new GL_RUN();
  soft_obj   = 0;// Emiliano
  irun = -1;
  runfirstentry = 0ULL;
  runlastentry = 0ULL;    
  //
};

Bool_t PamLevel2::IsGood(){
  Bool_t goodev=true;
  //  if(trk2_obj && trk2_obj->UnpackError() != 0 ) goodev = false; 
  if(calo2_obj && calo2_obj->good != 1) goodev = false;
  if(tof_obj && tof_obj->unpackError != 0) goodev = false;  
  if(trig_obj && trig_obj->unpackError != 0) goodev = false; 
  if(s4_obj && s4_obj->unpackError != 0) goodev = false;   
  if(nd_obj && nd_obj->unpackError != 0) goodev = false;   
  if(ac_obj && ac_obj->unpackError != 255) goodev = false;   
    //  if(orb_obj)  
  return goodev;
};

//--------------------------------------
//
//
//--------------------------------------
void *PamLevel2::GetPointerTo(const char* c ){

    TString objname = c;

    if(!objname.CompareTo("TrkLevel1"))  {
	if(!trk1_obj){ 
	    trk1_obj  = new TrkLevel1();
	    trk1_obj->Set();
	}
	return &trk1_obj;
    };
    if(!objname.CompareTo("TrkLevel2"))  {
	if(!trk2_obj){  
	    trk2_obj  = new TrkLevel2();
	    trk2_obj->Set();
	}
	return &trk2_obj;
    };
    if(!objname.CompareTo("TrkHough"))   {
	if(!trkh_obj)  trkh_obj  = new TrkHough();
	return &trkh_obj;
    };
    if(!objname.CompareTo("CaloLevel1")) {
	if(!calo1_obj) calo1_obj = new CaloLevel1();
	return &calo1_obj;
    };
    if(!objname.CompareTo("CaloLevel2")) {
	if(!calo2_obj){
	    calo2_obj = new CaloLevel2();
	    calo2_obj->Set();
	};
	return &calo2_obj;
    };
    if(!objname.CompareTo("ToFLevel2"))  {
	if(!tof_obj){
	    tof_obj   = new ToFLevel2();
	    tof_obj->Set();
	}
	return &tof_obj;
    };
    if(!objname.CompareTo("TrigLevel2")) {
	if(!trig_obj)  trig_obj  = new TrigLevel2();
	return &trig_obj;
    };
    if(!objname.CompareTo("S4Level2"))   {
	if(!s4_obj)    s4_obj    = new S4Level2();
	return &s4_obj;
    };
    if(!objname.CompareTo("NDLevel2"))   {
	if(!nd_obj)    nd_obj    = new NDLevel2();
	return &nd_obj;
    };
    if(!objname.CompareTo("AcLevel2"))   {
	if(!ac_obj)    ac_obj    = new AcLevel2();
	return &ac_obj;
    };
    if(!objname.CompareTo("OrbitalInfo")){
	if(!orb_obj)   orb_obj   = new OrbitalInfo();
	return &orb_obj;
    };
    
    if(!objname.CompareTo("RunInfo"))return &run_obj;

    if(!objname.CompareTo("SoftInfo"))return &soft_obj; // Emiliano

    return NULL;
};
//--------------------------------------
//
//
//--------------------------------------
/**
 * Retrieves the calorimeter track matching the seqno-th tracker stored track. 
 * (If seqno = -1 retrieves the self-trigger calorimeter track)
 */
 CaloTrkVar *PamLevel2::GetCaloStoredTrack(int seqno){
     
     if( !calo2_obj )return 0;

     if( calo2_obj->CaloLevel2::ntrk()==0 ){
	 cout << "PamLevel2::GetCaloStoredTrack(int) : requested tracker SeqNo "<< seqno <<" but no Calorimeter tracks are stored"<<endl;
	 return NULL;
     };
     
     CaloTrkVar *c = 0;
     Int_t it_calo=0;
     
     do{ 
	 c = calo2_obj->CaloLevel2::GetCaloTrkVar(it_calo);
	 it_calo++; 
     } while( c && seqno != c->trkseqno && it_calo < calo2_obj->CaloLevel2::ntrk());	
     
     if(!c || seqno != c->trkseqno){
	 c = 0;
	 if(seqno!=-1)cout << "PamLevel2::GetCaloStoredTrack(int) : requested tracker SeqNo "<< seqno <<" does not match Calorimeter stored tracks"<<endl;
     };
     return c;
     
 };
//--------------------------------------
 //
 //
//--------------------------------------
/**
  * Retrieves the ToF track matching the seqno-th tracker stored track. 
  * (If seqno = -1 retrieves the tracker-independent tof track)
 */
 ToFTrkVar *PamLevel2::GetToFStoredTrack(int seqno){
	 
     if( !tof_obj )return 0;

     if( tof_obj->ToFLevel2::ntrk()==0 ){
	 cout << "PamLevel2::GetToFStoredTrack(int) : requested tracker SeqNo "<< seqno <<" but no ToF tracks are stored"<<endl;
	 return NULL;
     };
     
     ToFTrkVar *c = 0;
     Int_t it_tof=0;
     
     do{ 
	 c = tof_obj->ToFLevel2::GetToFTrkVar(it_tof);
	 it_tof++; 
     } while( c && seqno != c->trkseqno && it_tof < tof_obj->ToFLevel2::ntrk());	
     
     if(!c || seqno != c->trkseqno){
	 c = 0;
	 if(seqno!=-1)cout << "PamLevel2::GetToFStoredTrack(int) : requested tracker SeqNo "<< seqno <<" does not match ToF stored tracks"<<endl;
     };
     return c;
     
 };

//--------------------------------------
 //
 //
//--------------------------------------
/**
 * Give the pamela track associated to a tracker track, retrieving related calorimeter and tof track information. 
 */
 PamTrack* PamLevel2::GetPamTrackAlong(TrkTrack* t){
     
     cout <<"PamLevel2::GetPamTrackAlong(TrkTrack* t) **obsolete** "<<endl;
     cout <<"(if you use it, remember to delete the PamTrack object)"<<endl;

     CaloTrkVar *c = 0;
     ToFTrkVar  *o = 0;
     
     if(CAL2) c = GetCaloStoredTrack(t->GetSeqNo());
     if(TOF) o = GetToFStoredTrack(t->GetSeqNo());
     
//    if(t && c && o)track = new PamTrack(t,c,o);
     PamTrack *track = new PamTrack(t,c,o);
     
     return track;
 
 };
//--------------------------------------
//
//
//--------------------------------------
/**
 * Retrieves the it-th stored track. 
 * It override TrkLevel2::GetTrack(int it). 
 * @param itrk Track number, ranging from 0 to GetNTracks().
 */

PamTrack* PamLevel2::GetStoredTrack(Int_t itrk){
    
    cout <<"PamLevel2::GetStoredTrack(Int_t itrk) **to-be-updated** "<<endl;
    cout <<"for the moment, better use separately the methods: TrkLevel2::GetStoredTrack(seqno) CaloLevel2::GetCaloTrkVar(Int_t notrack) ToFLevel2::GetToFTrkVar(Int_t notrack)"<<endl;
    cout <<"(if you use it, remember to delete the PamTrack object)"<<endl;
    PamTrack *track = 0;
    
    if( itrk >=0 && itrk < trk2_obj->TrkLevel2::ntrk() ){
	
	TrkTrack *t = trk2_obj->TrkLevel2::GetStoredTrack(itrk);
	track = GetPamTrackAlong(t);
	
	

    }else{
	cout << "PamLevel2::GetStoredTrack(int) : tracker track SeqNo "<< itrk <<" does not exist (GetNTracks() = "<<trk2_obj->TrkLevel2::GetNTracks()<<")"<<endl;
    };
    
    return track;
    
}
//--------------------------------------
//

/**
 * Sort physical (tracker) tracks and stores them in the TRefArray (of TrkTrack objects) which pointer is  PamLevel2::sorted_tracks. Left here as backward compatibility method.
 **/
void PamLevel2::SortTracks(TString how){
  printf(" WARNING! obsolete, use SortTracks() and SetSortingMethod(TString) instead! \n Setting sorting method to %s \n",how.Data());
  howtosort = how;  
  SortTracks();
};

//
//--------------------------------------
/**
 * Sort physical (tracker) tracks and stores them in the TRefArray (of TrkTrack objects) which pointer is  PamLevel2::sorted_tracks.
 * @param how String to set the sorting cryterium (es: "CAL" or "TRK+CAL+TOF" ecc...).
 * Sorting cryteria:
 * TRK: lower chi**2 
 * CAL: lower Y spatial residual on the first calorimeter plane
 * TOF: bigger numebr of hit PMTs along the track, on S12 S21 S32 (where paddles are along the Y axis).
 * The default sorting cryterium is "TOF+CAL".
 * 
 * The total number of physical tracks is always given by GetNTracks() and the it-th physical track can be retrieved by means of the methods GetTrack(int it) and GetTrack(int it, TString how).
 */
void PamLevel2::SortTracks(){
  TString how = howtosort;

  //    cout <<" PamLevel2::SortTracks(TString how) "<<endl;
  if( !trk2_obj ){
    cout << "void PamLevel2::SortTracks():  TrkLevel2 not loaded !!!";
    return;
  };
  //    cout << "call SortTracs() "<<endl;
  //Save current Object count
  Int_t ObjectNumber = TProcessID::GetObjectCount();

  //    cout << "ObjectNumber  "<<ObjectNumber <<endl;

  //     if(!sorted_tracks)sorted_tracks = new TRefArray();
  //     sorted_tracks->Clear();
  //    sorted_tracks.Clear();

  if(!tsorted)tsorted = new TClonesArray("PamTrack",trk2_obj->GetNTracks());
  tsorted->Delete();
  TClonesArray &ttsorted = *tsorted;
  if(!timage)timage = new TClonesArray("PamTrack",trk2_obj->GetNTracks());
  timage->Delete();
  TClonesArray &ttimage = *timage;



  //--------------------------------------------------
  // retrieve sorting method
  //--------------------------------------------------
  Bool_t use_TRK = how.Contains("TRK", TString::kIgnoreCase);
  Bool_t use_CAL = how.Contains("CAL", TString::kIgnoreCase);
  Bool_t use_TOF = how.Contains("TOF", TString::kIgnoreCase);
    
  if( !CAL2 &&  use_CAL) use_CAL = false;
  if( !TOF &&  use_TOF) use_TOF = false;
    
  if( !TRK2 ){
    //	cout << "SortTracks() : without tracker does not work!!! (not yet)" << endl;
    return;
  };

  //   cout << "use_CAL "<<use_CAL<<" use_TOF "<<use_TOF<<" use_TRK "<<use_TRK <<endl;
  

  //--------------------------------------------------
  // loop over "physical" tracks sorted by the tracker
  //--------------------------------------------------
  for(Int_t i=0; i < trk2_obj->TrkLevel2::GetNTracks(); i++){
	
    TrkTrack   *ts = 0;
    CaloTrkVar *cs = 0;
    ToFTrkVar  *os = 0;
	
    // get tracker tracks
    TrkTrack   *tp = trk2_obj->TrkLevel2::GetTrack(i); //tracker
    CaloTrkVar *cp = GetCaloStoredTrack(tp->GetSeqNo());
    ToFTrkVar  *op = GetToFStoredTrack(tp->GetSeqNo());

    TrkTrack   *ti = 0; //tracker (image)
    CaloTrkVar *ci = 0;
    ToFTrkVar  *oi = 0;
    //	cout << "trk track n. "<<i << " "<<hex<< tp <<dec<< endl;
    // if track has an image, check image selection
    if(tp->HasImage()){
	    
      ti = trk2_obj->TrkLevel2::GetTrackImage(i);              //tracker (image)
      ci = GetCaloStoredTrack(ti->GetSeqNo());
      oi = GetToFStoredTrack(ti->GetSeqNo());
	    
      //	    cout << "its image "<<i << " "<<hex<< ti <<dec<< endl;

      //assign starting scores
      Int_t tp_score = 0;  //main track sorted by the tracker
      Int_t ti_score = 0;  //image track
	    
      // -----------------------------------------------------------------------------------------
      // calorimeter check
      // -----------------------------------------------------------------------------------------
      // check the Y spatial residual on the first calorimeter plane
      // (cut on calorimeter variables from Emiliano)
      if( use_CAL && !calo2_obj ){
	cout << "void PamLevel2::SortTracks(): howtosort= "<<how<<" but CaloLevel2 not loaded !!!";
	return;
      };
      if(
	 use_CAL            &&
//	 calo2_obj->npcfit[1] > 5     &&   //no. of fit planes on Y view
//	 calo2_obj->varcfit[1] < 1000. &&  //fit variance on Y view
	 cp && ci &&
	 true){

		
// 	Float_t resy_p = cp->tbar[0][1] - calo2_obj->cbar[0][1]; if(resy_p < 0)resy_p= - resy_p;
// 	Float_t resy_i = ci->tbar[0][1] - calo2_obj->cbar[0][1]; if(resy_i < 0)resy_i= - resy_i;
		
// 	if(resy_p <= resy_i) tp_score++;
// 	else                 ti_score++;
	

	  if( cp->npresh > ci->npresh) tp_score++;
	  if( cp->npresh < ci->npresh) ti_score++;
	  else ;//niente

	//	cout << "CALO "<<tp_score<<ti_score<<endl;

      };
      // -----------------------------------------------------------------------------------------
      // TOF check
      // -----------------------------------------------------------------------------------------
      // check the number of hit pmts along the track 
      // on S12 S21 and S32, where paddles are parallel to Y axis
      if( use_TOF && !tof_obj ){
	cout << "void PamLevel2::SortTracks(): howtosort= "<<how<<" but ToFLevel2 not loaded !!!";
	return;
      };
      if( use_TOF && op && oi ){
		
	//
	Float_t sen = 0.;
	for (Int_t ih=0; ih < op->npmtadc; ih++){
	  Int_t pl = tof_obj->GetPlaneIndex( (op->pmtadc).At(ih) );
	  if ( pl == 2 || pl == 3 || pl == 4 || pl == 5 ) sen += (op->dedx).At(ih);
	};
	for (Int_t ih=0; ih < oi->npmtadc; ih++){
	  Int_t pl = tof_obj->GetPlaneIndex( (oi->pmtadc).At(ih) );
	  if ( pl == 2 || pl == 3 || pl == 4 || pl == 5 ) sen += (oi->dedx).At(ih);
	};
	//
	if (  sen >= sortthr ){
	  //printf(" IS A NUCLEUS! en = %f \n",sen);
	  //
	  // is a nucleus use a different algorithm
	  //
	  Int_t nz = 6; Float_t zin[6];                                          // << define TOF z-coordinates
	  for(Int_t ip=0; ip<nz; ip++) 
	    zin[ip] = tof_obj->ToFLevel2::GetZTOF(tof_obj->ToFLevel2::GetToFPlaneID(ip));     // << read ToF plane z-coordinates
	  Trajectory *tr = new Trajectory(nz,zin);
	  //
	  Int_t nphit_p =0;
	  Int_t nphit_i =0;
	  Float_t enhit_p = 0.;
	  Float_t enhit_i = 0.;
	  //
	  for (Int_t ih=0; ih < op->npmtadc; ih++){
	    Int_t pl = tof_obj->GetPlaneIndex( (op->pmtadc).At(ih) );
	    if(pl == 1 || pl == 2 || pl == 5){
	      nphit_p++;
	      enhit_p += (op->dedx).At(ih);
	    };
	  };
	  //
	  tp->DoTrack2(tr);
	  //
	  if ( fabs(tr->y[0]-oi->ytofpos[0]) < 2. ){
	    for (Int_t ih=0; ih < op->npmtadc; ih++){
	      Int_t pl = tof_obj->GetPlaneIndex( (op->pmtadc).At(ih) );
	      if(pl == 0){
		nphit_p++;
		enhit_p += (op->dedx).At(ih);
	      };
	    };
	  };
	  if ( fabs(tr->y[3]-oi->ytofpos[1]) < 2. ){
	    for (Int_t ih=0; ih < op->npmtadc; ih++){
	      Int_t pl = tof_obj->GetPlaneIndex( (op->pmtadc).At(ih) );
	      if(pl == 3){
		nphit_p++;
		enhit_p += (op->dedx).At(ih);
	      };
	    };
	  };
	  if ( fabs(tr->y[4]-oi->ytofpos[2]) < 2. ){
	    for (Int_t ih=0; ih < op->npmtadc; ih++){
	      Int_t pl = tof_obj->GetPlaneIndex( (op->pmtadc).At(ih) );
	      if(pl == 4){
		nphit_p++;
		enhit_p += (op->dedx).At(ih);
	      };
	    };
	  };

	  for (Int_t ih=0; ih < oi->npmtadc; ih++){
	    Int_t pl = tof_obj->GetPlaneIndex( (oi->pmtadc).At(ih) );
	    if(pl == 1 || pl == 2 || pl == 5){
	      nphit_i++;	
	      enhit_i += (op->dedx).At(ih);
	    };
	  };
	  //
	  ti->DoTrack2(tr);
	  //
	  if ( fabs(tr->y[0]-oi->ytofpos[0]) < 2. ){
	    for (Int_t ih=0; ih < oi->npmtadc; ih++){
	      Int_t pl = tof_obj->GetPlaneIndex( (oi->pmtadc).At(ih) );
	      if(pl == 0){
		nphit_i++;
		enhit_i += (op->dedx).At(ih);
	      };
	    };
	  };
	  if ( fabs(tr->y[3]-oi->ytofpos[1]) < 2. ){
	    for (Int_t ih=0; ih < oi->npmtadc; ih++){
	      Int_t pl = tof_obj->GetPlaneIndex( (oi->pmtadc).At(ih) );
	      if(pl == 3){
		nphit_i++;
		enhit_i += (op->dedx).At(ih);
	      };
	    };
	  };
	  if ( fabs(tr->y[4]-oi->ytofpos[2]) < 2. ){
	    for (Int_t ih=0; ih < oi->npmtadc; ih++){
	      Int_t pl = tof_obj->GetPlaneIndex( (oi->pmtadc).At(ih) );
	      if(pl == 4){
		nphit_i++;
		enhit_i += (op->dedx).At(ih);
	      };
	    };
	  };		

		
	  if(
	     use_TOF            &&
	     (nphit_p+nphit_i) !=0 &&	
	     true){
		    
	    //	    printf(" seqno %i nphit_p %i nphit_i %i enhit_p %f enhit_i %f \n",trk2_obj->TrkLevel2::GetSeqNo(i),nphit_p,nphit_i,enhit_p,enhit_i); 
	    //	    printf(" score p %i score i %i \n",tp_score,ti_score);
	    //		  if( enhit_p > enhit_i ) tp_score++;
	    //		  if( nphit_p >= nphit_i && enhit_p > enhit_i ) tp_score++;
	    if ( nphit_p > nphit_i ) tp_score++;
	    if ( nphit_p < nphit_i ) ti_score++;
	    if ( nphit_p == nphit_i ){
	      if ( enhit_p > enhit_i ) tp_score++;
	      else ti_score++;
	    };
	    //	    printf(" dopo score p %i score i %i \n",tp_score,ti_score);
	  };
	  delete tr;
	  //
	} else {
	  //
	  // NOT a NUCLEUS
	  //
	  //printf(" NOT a NUCLEUS! en = %f \n",sen);

	  Int_t nphit_p =0;
	  Int_t nphit_i =0;
		
		
	  /*				cout << "track: npmtadc "<< op->npmtadc << endl;
	    cout << "track: npmttdc "<< op->npmttdc << endl;
	    cout << "image: npmtadc "<< oi->npmtadc << endl;
	    cout << "image: npmttdc "<< oi->npmttdc << endl;*/
		
// 	  for (Int_t ih=0; ih < op->npmtadc; ih++){
// 	    Int_t pl = tof_obj->GetPlaneIndex( (op->pmtadc).At(ih) );
// 	    if(pl == 1 || pl == 2 || pl == 5)nphit_p++;
// 	  };
		
// 	  for (Int_t ih=0; ih < oi->npmtadc; ih++){
// 	    Int_t pl = tof_obj->GetPlaneIndex( (oi->pmtadc).At(ih) );
// 	    if(pl == 1 || pl == 2 || pl == 5)nphit_i++;	
// 	  };
	  // --- modified to count tdc signals (more efficient?)
	  // --- and to implement check on tdcflag
	  for (Int_t ih=0; ih < op->npmttdc; ih++){
	    //	    Int_t pl = tof_obj->GetPlaneIndex( (op->pmttdc).At(ih) );
//	    if( (op->tdcflag).At(ih)==0 && (pl == 1 || pl == 2 || pl == 5) )nphit_p++;
	    if( (op->tdcflag).At(ih)==0 )nphit_p++;
	  };
		
	  for (Int_t ih=0; ih < oi->npmttdc; ih++){
	    //	    Int_t pl = tof_obj->GetPlaneIndex( (oi->pmttdc).At(ih) );
//	    if( (oi->tdcflag).At(ih)==0 && (pl == 1 || pl == 2 || pl == 5) )nphit_i++;	
	    if( (oi->tdcflag).At(ih)==0 )nphit_i++;	
	  };
		
	  if(
	     (nphit_p+nphit_i) !=0 &&	
	     true){
		    
	    if     ( nphit_p > nphit_i) tp_score++;
	    else if( nphit_p < nphit_i) ti_score++;
	    else ;//niente
	  };
	};
	//	cout << "TOF "<<tp_score<<ti_score<<endl;
      };


//      if(tp_score == ti_score) use_TRK = true;


      // -----------------------------------------------------------------------------------------
      // tracker check
      // -----------------------------------------------------------------------------------------
      // chi**2 difference is not always large enough to distinguish among 
      // the real track and its image.
      // Tracker check will be applied always when calorimeter and tof information is ambiguous.
      // *** MODIFIED ON AUGUST 2007 ***
      if(use_TRK){
// 	if(      tp->chi2 > 0 && tp->chi2 < ti->chi2 ) tp_score++ ;
// 	else if( ti->chi2 > 0 && ti->chi2 < tp->chi2 ) ti_score++ ;

	  // CHECK 1 : number of points along X  
	  if ( tp->GetNX() >= ti->GetNX() )tp_score++ ;
	  if ( tp->GetNX() <= ti->GetNX() )ti_score++ ;
	  // CHECK 2 : number of points along Y  
	  if ( tp->GetNY() >= ti->GetNY() )tp_score++ ;
	  if ( tp->GetNY() <= ti->GetNY() )ti_score++ ;
	  // CHECK 3 : chi**2 along X  
// 	  if(      tp->GetChi2X() > 0 && (tp->GetChi2X() < ti->GetChi2X() || ti->GetChi2X() <=0) ) tp_score++ ;
// 	  if(      ti->GetChi2X() > 0 && (ti->GetChi2X() < tp->GetChi2X() || tp->GetChi2X() <=0) ) ti_score++ ;
	  // CHECK 4 : chi**2 along Y  
// 	  if(      tp->GetChi2Y() > 0 && (tp->GetChi2Y() < ti->GetChi2Y() || ti->GetChi2Y() <=0) ) tp_score++ ;
// 	  if(      ti->GetChi2Y() > 0 && (ti->GetChi2Y() < tp->GetChi2Y() || tp->GetChi2Y() <=0) ) ti_score++ ;
	  // CHECK 5 : total chi**2   
// 	  if(      tp->chi2 > 0 && (tp->chi2 < ti->chi2 || ti->chi2 <=0) ) tp_score++ ;
// 	  if(      ti->chi2 > 0 && (ti->chi2 < tp->chi2 || tp->chi2 <=0) ) ti_score++ ;

	//	cout << "TRK "<<tp_score<<ti_score<<endl;
      };
	    


      // *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
      // the winner is....
      // *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
      if      (tp_score > ti_score) {
	 

      }else if (tp_score < ti_score) {


	ts = ti;//its image!!
	cs = ci;
	os = oi;

	ti = tp;//its image!!
	ci = cp;
	oi = op;

	tp = ts;//its image!!
	cp = cs;
	op = os;

		
      }else {

//	cout << "Warning - track image ambiguity not solved" << endl;

      };
	    
    }else{
      // 	    ts = tp;
      // 	    cs = cp;
      // 	    os = op;
    };
	
    //	cout <<" SortTracks() "<<i<<" -- "<<ts<<endl;
    //	sorted_tracks->Add(ts);//save the track in the sorted array
    //	sorted_tracks.Add(ts);//save the track in the sorted array
    //	sorted_tracks.Add(tp);//save the track in the sorted array
    //	cout << "SortTracks:: sorted_tracks->Add(it) "<<i<<" "<<ts<<endl;
    //	cout<<"o "<<tp<<endl;
    //	cout<<"o "<<cp<<endl;
    //	cout<<"o "<<op<<endl;
    new(ttsorted[i]) PamTrack(tp,cp,op);
    new(ttimage[i])  PamTrack(ti,ci,oi);
  };

  if( tsorted->GetEntries() != trk2_obj->GetNTracks() ){
    cout << "void PamLevel2::SortTracks(): tsorted->GetEntries() "<<tsorted->GetEntries()<<" != trk2_obj->GetNTracks() = "<<trk2_obj->GetNTracks() <<endl;
    tsorted->Delete(); tsorted=0;
    timage->Delete(); timage=0;	
  }

  //Restore Object count 
  //To save space in the table keeping track of all referenced objects
  //We reset the object count to what it was at the beginning of the event.
  TProcessID::SetObjectCount(ObjectNumber);
    
};
//--------------------------------------
//
//
//--------------------------------------
/**
 * This method overrides TrkLevel2::GetTracks(), where sorting is done by decreasing number of fit points and increasing chi^2.
 * PamLevel2::GetTracks() keeps the same track order given by TrkLevel2::GetTracks(), but checks image selection by using calorimeter and ToF tracking information.
 */
// TRefArray *PamLevel2::GetTracks(){

// //  *-*-*-*-*-*-*-*-*-*-*-*-*
//     SortTracks("+CAL+TOF");
// //  *-*-*-*-*-*-*-*-*-*-*-*-*

// //   return  sorted_tracks;
//     return &sorted_tracks;
    
//  };
TClonesArray *PamLevel2::GetTracks(){

//  *-*-*-*-*-*-*-*-*-*-*-*-*
    SortTracks();
//  *-*-*-*-*-*-*-*-*-*-*-*-*

    return tsorted;
    
 };
//--------------------------------------
 //
 //
//--------------------------------------
/**
 * Retrieves the it-th Pamela "physical" track. 
 * It override TrkLevel2::GetTrack(int it). 
 * @param it Track number, ranging from 0 to GetNTracks().
 */
PamTrack *PamLevel2::GetTrack(int it){

//    if(!trk2_obj) return 0;
	
// //  *-*-*-*+-*-*-*-*-*-*-*-*-*
//     SortTracks("+CAL+TOF");
// //  *-*-*-*-*-*-*-*-*-*-*-*-*
// //    if(!sorted_tracks)return 0;
//     if(sorted_tracks.IsEmpty())return 0;

//     PamTrack *track = 0;
    
//     if( it >=0 && it < trk2_obj->TrkLevel2::GetNTracks() && it<sorted_tracks.GetEntriesFast() ){	
// 	TrkTrack   *t = (TrkTrack*)sorted_tracks.At(it);
// 	track = GetPamTrackAlong(t);
//     }else{
// 	cout << "PamLevel2::GetTrack(int) : tracker track SeqNo "<< it <<" does not exist (GetNTracks() = "<<trk2_obj->TrkLevel2::GetNTracks()<<")"<<endl;
//     };
    
//     return track;

//    cout << "PamLevel2::GetTrack(int it) "<<endl;
//  *-*-*-*-*-*-*-*-*-*-*-*-*
    SortTracks();
//  *-*-*-*-*-*-*-*-*-*-*-*-*
    if(!tsorted)return 0;
    if(!tsorted->GetEntries())return 0;

    PamTrack *track = 0;
    
    if( it >=0 && it < trk2_obj->TrkLevel2::GetNTracks()){
// 	TrkTrack   *t = (TrkTrack*)sorted_tracks.At(it);
// 	track = GetPamTrackAlong(t);
 	TClonesArray &t = *(tsorted);
 	track = (PamTrack*)t[it];
	
    }else{
	cout << "PamLevel2::GetTrack(int) : tracker track SeqNo "<< it <<" does not exist (GetNTracks() = "<<trk2_obj->TrkLevel2::GetNTracks()<<")"<<endl;
    };
    
    return track;
    
};

//--------------------------------------
//
//
//--------------------------------------
/**
 * Retrieves (if present) the image of the it-th Pamela "physical" track, sorted by the method PamLevel2::SortTracks(). 
 * @param it Track number, ranging from 0 to GetNTracks().
 */
PamTrack *PamLevel2::GetTrackImage(int it){

//     if(!trk2_obj) return 0;

// //  *-*-*-*-*-*-*-*-*-*-*-*-*
//     SortTracks("+CAL+TOF");
// //  *-*-*-*-*-*-*-*-*-*-*-*-*
// //    if(!sorted_tracks)return 0;
//     if(sorted_tracks.IsEmpty())return 0;
    
//     PamTrack *image = 0;
    
//     if( it >=0 && it < trk2_obj->TrkLevel2::GetNTracks() ){
// 	TrkTrack *temp = (TrkTrack*)sorted_tracks.At(it);
// 	if( temp->HasImage() ){
// 	    TrkTrack   *t = trk2_obj->TrkLevel2::GetStoredTrack(temp->GetImageSeqNo());
// 	    image = GetPamTrackAlong(t);
// 	}else{
// 	    cout <<"PamLevel2::GetTrackImage(int) : Track SeqNo "<<it<<" does not have image"<<endl;
// 	};
//     }else{
// 	cout << "PamLevel2::GetTrackImage(int) : Tracker track SeqNo "<< it <<" does not exist (GetNTracks() = "<<trk2_obj->TrkLevel2::GetNTracks()<<")"<<endl;
//     };
    
//     return image;


//  *-*-*-*-*-*-*-*-*-*-*-*-*
    SortTracks();
//  *-*-*-*-*-*-*-*-*-*-*-*-*
    if(!timage)return 0;
    if(!timage->GetEntries())return 0;
    
    PamTrack *image = 0;
    
    if( it >=0 && it < trk2_obj->TrkLevel2::GetNTracks() ){
 	TClonesArray &t = *(tsorted);
 	PamTrack *temp = (PamTrack*)t[it];
	if( temp->GetTrkTrack()->HasImage() ){
	    TClonesArray &t = *(timage);
	    image = (PamTrack*)t[it];
	}else{
//	    cout <<"PamLevel2::GetTrackImage(int) : Track SeqNo "<<it<<" does not have image"<<endl;
	};
    }else{
	cout << "PamLevel2::GetTrackImage(int) : Tracker track SeqNo "<< it <<" does not exist (GetNTracks() = "<<trk2_obj->TrkLevel2::GetNTracks()<<")"<<endl;
    };
    
    return image;
}

//--------------------------------------
//
//
//--------------------------------------
/**
 * Get the Pamela detector trees in a single file and make them friends.
 * @param f TFile pointer
 * @param detlist String to select trees to be included
 * @return Pointer to a TTree
 */
TTree *PamLevel2::GetPamTree(TFile *f, TString detlist ){

//     if( !detlist.IsNull() ) SetWhichTrees(detlist);
//     else                    GetWhichTrees(f);
  if ( pam_tree ){
    printf("WARNING: TTree *PamLevel2::GetPamTree(TFile *fl, TString detlist) -- pam_tree already exists!\n ");
    return pam_tree;
  };
  //

    cout << "TTree *PamLevel2::GetPamTree(TFile *f, TString detlist ) -- obsolte "<<endl;

    SetWhichTrees(detlist);

    TTree *Trout =0;

    TString fname = f->GetName();
    if(!CheckLevel2File(fname))return NULL;

//    UInt_t *found=0;    

    cout<< "GetPamTree(TFile*,TString): detector list --> ";
    if(TRK1)cout<<"TRK1 ";
    if(TRK2)cout<<"TRK2 ";
    if(TRKh)cout<<"TRKH ";
    if(CAL1)cout<<"CAL1 ";
    if(CAL2)cout<<"CAL2 ";
    if(TOF)cout<<"TOF ";
    if(TRG)cout<<"TRG ";
    if(AC)cout<<"AC ";
    if(ND)cout<<"ND ";
    if(S4)cout<<"S4 ";
    if(ORB)cout<<"ORB ";
    cout << endl;
    if(SELLI)cout<<">>> Found selection-list <<<"<<endl;

    f->cd();

// Tracker
    TTree *T = (TTree*)f->Get("Tracker");
    if(T && (TRK2||TRK1||TRKh)) {
	if(TRK2)T->SetBranchAddress("TrkLevel2", GetPointerTo("TrkLevel2"));
//	else    T->SetBranchStatus("TrkLevel2",0,found);
	if(TRK2)cout << "Tracker      : set branch address TrkLevel2"<<endl;
	if(TRK1)T->SetBranchAddress("TrkLevel1", GetPointerTo("TrkLevel1"));
//	else    T->SetBranchStatus("TrkLevel1",0,found);
	if(TRK1)cout << "Tracker      : set branch address TrkLevel1"<<endl;
 	if(TRKh)T->SetBranchAddress("TrkHough", GetPointerTo("TrkHough"));
//	else    T->SetBranchStatus("TrkHough",0,found);
 	if(TRKh)cout << "Tracker      : set branch address TrkHough"<<endl;
        if(!Trout)Trout=T;
        else Trout->AddFriend(T);
    }else{
        cout << "Tracker      : missing tree"<<endl;
    };
    // Calorimeter
    TTree *C = (TTree*)f->Get("Calorimeter");
    if(C && (CAL2||CAL1)) {
        if(CAL2)C->SetBranchAddress("CaloLevel2", GetPointerTo("CaloLevel2"));
//	else    C->SetBranchStatus("CaloLevel2",0,found);
        if(CAL2)cout << "Calorimeter  : set branch address CaloLevel2"<<endl;
        if(CAL1)C->SetBranchAddress("CaloLevel1", GetPointerTo("CaloLevel1"));
//	else    C->SetBranchStatus("CaloLevel1",0,found);
        if(CAL1)cout << "Calorimeter  : set branch address CaloLevel1"<<endl;
        if(!Trout)Trout=C;
        else Trout->AddFriend(C);
    }else{
        cout << "Calorimeter  : missing tree"<<endl;
    };

    // ToF    
    TTree *O = (TTree*)f->Get("ToF");
    if(O && TOF) {
        O->SetBranchAddress("ToFLevel2", GetPointerTo("ToFLevel2"));
        cout << "ToF          : set branch address ToFLevel2"<<endl;
        if(!Trout)Trout=O;
        else Trout->AddFriend(O);
    }else{
        cout << "ToF         : missing tree"<<endl;
    };
    // Trigger
    TTree *R = (TTree*)f->Get("Trigger");
    if(R && TRG) {
        R->SetBranchAddress("TrigLevel2", GetPointerTo("TrigLevel2"));
        cout << "Trigger      : set branch address TrigLevel2"<<endl;
        if(!Trout)Trout=R;
        else Trout->AddFriend(R);
    }else{
        cout << "Trigger      : missing tree"<<endl;
    };
    // S4
    TTree *S = (TTree*)f->Get("S4");
    if(S && S4) {
        S->SetBranchAddress("S4Level2", GetPointerTo("S4Level2"));
        cout << "S4           : set branch address S4Level2"<<endl;
        if(!Trout)Trout=S;
        else Trout->AddFriend(S);
    }else{
        cout << "S4           : missing tree"<<endl;
    };
    // Neutron Detector
    TTree *N = (TTree*)f->Get("NeutronD");
    if(N && ND) {
        N->SetBranchAddress("NDLevel2", GetPointerTo("NDLevel2"));
        cout << "NeutronD     : set branch address NDLevel2"<<endl;
        if(!Trout)Trout=N;
        else Trout->AddFriend(N);
    }else{
        cout << "NeutronD     : missing tree"<<endl;
    };
    // Anticounters
    TTree *A = (TTree*)f->Get("Anticounter");
    if(A && AC) {
        A->SetBranchAddress("AcLevel2", GetPointerTo("AcLevel2"));
        cout << "Anticounter  : set branch address AcLevel2"<<endl;
        if(!Trout)Trout=A;
        else Trout->AddFriend(A);
    }else{
        cout << "Anticounter  : missing tree"<<endl;
    };
    // Orbital Info
    TTree *B = (TTree*)f->Get("OrbitalInfo");
    if(B && ORB) {
	B->SetBranchAddress("OrbitalInfo", GetPointerTo("OrbitalInfo"));
	cout << "OrbitalInfo  : set branch address OrbitalInfo"<<endl;
	if(!Trout)Trout=B;
	else Trout->AddFriend(B);
    }else{
	cout << "OrbitalInfo  : missing tree"<<endl;
    };

    TTree *L = (TTree*)f->Get("SelectionList");
    if(L && SELLI==1) {
// 	L->SetBranchAddress("RunEntry",&irun);
// 	cout << "SelectionList: set branch address RunEntry"<<endl;
// 	L->SetBranchAddress("EventEntry",&irunentry);
// 	cout << "SelectionList: set branch address EventEntry"<<endl;
// 	if(!Trout)Trout=O;
// 	else Trout->AddFriend("SelectionList");
	cout << " TTree *PamLevel2::GetPamTree(TFile, TString) >>> SelectionList not implemente!!!!"<<endl;
	sel_tree = 0;
    }else{
	cout << "SelectionList  : missing tree"<<endl;
    };
    
    cout<<endl<<" Number of entries: "<<Trout->GetEntries()<<endl<<endl;

    pam_tree = (TChain*)Trout;
    
    return Trout;
    
}
//--------------------------------------
//
//
//--------------------------------------
/**
 * Get list of Level2 files.
 * @param ddir Level2 data directory.
 * @param flisttxt Name of txt file containing file list.
 * @return Pointer to a TList of TSystemFiles
 * If no input file list is given , all the Level2 files inside the directory are processed.
 */
TList*  PamLevel2::GetListOfLevel2Files(TString ddir, TString flisttxt = ""){
	
    TString wdir = gSystem->WorkingDirectory();	
    
    //    if(ddir=="")ddir = wdir;
//	TSystemDirectory *datadir = new TSystemDirectory(gSystem->BaseName(ddir),ddir);
    if ( ddir != ""){
      cout << "Level2 data directory : "<<  ddir << endl;
    } else {
      cout << "Level2 data directory not given as input: trying to evaluate from list or taking working directory " << endl;
    };
    TList *contents  = new TList; // create output list
    contents->SetOwner();
    
//    char *fullpath;
//    const char *fullpath;
    
    // if no input file list is given:	
    if ( flisttxt != "" ){
	
//	if( !gSystem->IsFileInIncludePath(flisttxt,&fullpath) ){ 	
// 	if( !(fullpath = gSystem->FindFile("./",flisttxt)) ){ 
// 	    cout <<"File "<<flisttxt<<" not found"<<endl;
// 	    return 0;
// 	}	
// 	flisttxt = fullpath;
      if ( !flisttxt.EndsWith(".root") ){

	flisttxt = gSystem->ConcatFileName(gSystem->DirName(flisttxt),gSystem->BaseName(flisttxt));

	if( !gSystem->ChangeDirectory(ddir) ){
	    cout << "Cannot change directory : "<<ddir<<endl;
	    return 0; 
	}
	
	cout <<"Input file list : " << flisttxt <<endl;
	ifstream in; 
	in.open(flisttxt, ios::in); //open input file list
	if(!in.good()){ 
	    cout <<" ERROR opening the file "<<endl;
	    gSystem->ChangeDirectory(wdir); // back to the working directory
 	    return 0;
 	}	
	int line=0;
	while (1) {
	    TString file;
	    in >> file;
	    if (!in.good()) break;
	    line++;
//	    cout <<"(1) " << file << endl;
	    if(file.IsNull()){
		cout << "-- list interrupted at line "<<line <<endl;
		break;
	    }
	    if(file.Contains("#"))file = file(0,file.First("#"));
//	    cout <<"(2) " << file << endl;
//	    if( gSystem->IsFileInIncludePath(file,&fullpath) ){
//	    if( (fullpath = gSystem->FindFile(ddir,file)) ){
	    if( file.EndsWith(".root") ){
	      TString filedir;
	      if (ddir != ""){
		filedir = ddir; // take the input dir
	      } else {
		gSystem->ChangeDirectory(wdir); // back to the working directory
		filedir = gSystem->DirName(file); // this will take the path if exist in the list otherwise it will return automatically the working dir
	      };
		char *fullpath = gSystem->ConcatFileName(gSystem->DirName(filedir),gSystem->BaseName(file));
		contents->Add(new TSystemFile(fullpath,gSystem->DirName(fullpath)));// add file to the list
		delete fullpath;
	    }
//	    }else{
//		if(file.Data()!="")cout << "File: "<<file<<" ---> missing "<< endl;
//	    };
	};
	in.close();
      } else {
	  if(flisttxt.Contains("#"))flisttxt = flisttxt(0,flisttxt.First("#"));
	  char *fullpath = gSystem->ConcatFileName(gSystem->DirName(flisttxt),gSystem->BaseName(flisttxt));
	  contents->Add(new TSystemFile(fullpath,gSystem->DirName(fullpath)));// add file to the list
	  delete fullpath;
      };	
    }else{
	
	cout << "No input file list given."<<endl;
	cout << "Check for existing root files."<<endl;
//		cout << "Warking directory: "<< gSystem->WorkingDirectory()<< endl;
	if ( ddir == "" ){
	  ddir = wdir;
	  cout << "Level2 data directory : "<<  ddir << endl;
	};

	TSystemDirectory *datadir = new TSystemDirectory(gSystem->BaseName(ddir),ddir);
	TList *temp = datadir->GetListOfFiles();
//		temp->Print();
//		cout << "*************************************" << endl;
	
	TIter next(temp);
	TSystemFile *questo = 0;
	
	
	while ( (questo = (TSystemFile*) next()) ) {
	    TString name =  questo-> GetName();
	    if( name.EndsWith(".root") ){
//		const char *fullpath = gSystem->FindFile(ddir,name);
//		char *fullpath;
//		gSystem->IsFileInIncludePath(name,&fullpath);
		char *fullpath = gSystem->ConcatFileName(gSystem->DirName(ddir),gSystem->BaseName(name));
		contents->Add(new TSystemFile(fullpath,gSystem->DirName(fullpath)));
		delete fullpath;
	    };
	}
	delete temp;
	delete datadir;
	
    };
    gSystem->ChangeDirectory(wdir); // back to the working directory
//	cout << endl << "Selected files:" << endl;
//	contents->Print();
    cout << contents->GetEntries()<<" files \n";
//	cout << endl;
//	cout << "Working directory: "<< gSystem->WorkingDirectory()<< endl;
    return contents;
};
//--------------------------------------
//
//
//--------------------------------------
/**
 * Get the Pamela detector chains from a list of files and make them friends.
 * @param fl Pointer to a TList of TSystemFiles 
 * @param detlist String to select trees to be included
 * @return Pointer to a TChain
 */
TChain *PamLevel2::GetPamTree(TList *fl, TString detlist ){
  //
  //
  //
  if ( pam_tree ){
    printf("WARNING: TChain *PamLevel2::GetPamTree(TList *fl, TString detlist) -- pam_tree already exists!\n ");
    return pam_tree;
  };
  //
    
    TChain *Trout =0;

//    if( !detlist.IsNull() )SetWhichTrees(detlist);
    SetWhichTrees(detlist);
    
    cout<< "GetPamTree(TList*,TString): input detector list --> ";
    if(TRK1)cout<<"TRK1 ";
    if(TRK2)cout<<"TRK2 ";
    if(TRKh)cout<<"TRKH ";
    if(CAL1)cout<<"CAL1 ";
    if(CAL2)cout<<"CAL2 ";
    if(TOF)cout<<"TOF ";
    if(TRG)cout<<"TRG ";
    if(AC)cout<<"AC ";
    if(ND)cout<<"ND ";
    if(S4)cout<<"S4 ";
    if(ORB)cout<<"ORB ";
    cout << endl;

     TChain *T = 0;	
     TChain *C = 0;
     TChain *O = 0;
     TChain *R = 0;
     TChain *S = 0;
     TChain *N = 0;
     TChain *A = 0;
     TChain *B = 0;

     TChain *L = 0;
    
    if(TRK2||TRK1||TRKh) T = new TChain("Tracker");	
    if(CAL2||CAL1)       C = new TChain("Calorimeter");
    if(TOF)              O = new TChain("ToF");
    if(TRG)              R = new TChain("Trigger");
    if(S4)               S = new TChain("S4");
    if(ND)               N = new TChain("NeutronD");
    if(AC)               A = new TChain("Anticounter");
    if(ORB)              B = new TChain("OrbitalInfo");
     L = new TChain("SelectionList");
    
    // loop over files and create chains	
    TIter next(fl);
    TSystemFile *questo = 0;
    while ( (questo = (TSystemFile*) next()) ) {
	TString name =  questo->GetName();
	cout << "File: "<< name << endl;
	if( CheckLevel2File(name) ){
	    if(TRK2||TRK1||TRKh) T->Add(name);
	    if(CAL1||CAL2)       C->Add(name);
	    if(TOF)              O->Add(name);
	    if(TRG)              R->Add(name);
	    if(S4)               S->Add(name);
	    if(ND)               N->Add(name);
	    if(AC)               A->Add(name);
	    if(ORB)              B->Add(name);
	    if(SELLI==1)         L->Add(name);
	};
    };
    
    cout << "done chain \n";
    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;

//    UInt_t *found=0;
// Tracker
    if(T && (TRK2||TRK1||TRKh)) {
	if(TRK2)T->SetBranchAddress("TrkLevel2", GetPointerTo("TrkLevel2"));
//	else    T->SetBranchStatus("TrkLevel2",0,found);
	if(TRK2)cout << "Tracker      : set branch address TrkLevel2"<<endl;
	if(TRK1)T->SetBranchAddress("TrkLevel1", GetPointerTo("TrkLevel1"));
//	else    T->SetBranchStatus("TrkLevel1",0,found);
	if(TRK1)cout << "Tracker      : set branch address TrkLevel1"<<endl;
 	if(TRKh)T->SetBranchAddress("TrkHough", GetPointerTo("TrkHough"));
//	else    T->SetBranchStatus("TrkHough",0,found);
 	if(TRKh)cout << "Tracker      : set branch address TrkHough"<<endl;
        if(!Trout)Trout=T;
	else Trout->AddFriend("Tracker");
    }else{
        cout << "Tracker      : missing tree"<<endl;
    };
    // Calorimeter
    if(C && (CAL2||CAL1)) {
        if(CAL2)C->SetBranchAddress("CaloLevel2", GetPointerTo("CaloLevel2"));
//	else    C->SetBranchStatus("CaloLevel2",0,found);
        if(CAL2)cout << "Calorimeter  : set branch address CaloLevel2"<<endl;
        if(CAL1)C->SetBranchAddress("CaloLevel1", GetPointerTo("CaloLevel1"));
//	else    C->SetBranchStatus("CaloLevel1",0,found);
        if(CAL1)cout << "Calorimeter  : set branch address CaloLevel1"<<endl;
        if(!Trout)Trout=C;
        else Trout->AddFriend("Calorimeter");
    }else{
        cout << "Calorimeter  : missing tree"<<endl;
    };
    // ToF    
    if(O && TOF) {
        O->SetBranchAddress("ToFLevel2", GetPointerTo("ToFLevel2"));
        cout << "ToF          : set branch address ToFLevel2"<<endl;
        if(!Trout)Trout=O;
        else Trout->AddFriend("ToF");
    }else{
        cout << "ToF         : missing tree"<<endl;
    };
    // Trigger
    if(R && TRG) {
        R->SetBranchAddress("TrigLevel2", GetPointerTo("TrigLevel2"));
        cout << "Trigger      : set branch address TrigLevel2"<<endl;
        if(!Trout)Trout=O;
        else Trout->AddFriend("Trigger");
    }else{
        cout << "Trigger      : missing tree"<<endl;
    };
    // S4
    if(S && S4) {
        S->SetBranchAddress("S4Level2", GetPointerTo("S4Level2"));
        cout << "S4           : set branch address S4Level2"<<endl;
        if(!Trout)Trout=O;
        else Trout->AddFriend("S4");
    }else{
        cout << "S4           : missing tree"<<endl;
    };
    // Neutron Detector
    if(N && ND) {
        N->SetBranchAddress("NDLevel2", GetPointerTo("NDLevel2"));
        cout << "NeutronD     : set branch address NDLevel2"<<endl;
        if(!Trout)Trout=O;
        else Trout->AddFriend("NeutronD");
    }else{
        cout << "NeutronD     : missing tree"<<endl;
    };
    // Anticounters
    if(A && AC) {
        A->SetBranchAddress("AcLevel2", GetPointerTo("AcLevel2"));
        cout << "Anticounter  : set branch address AcLevel2"<<endl;
        if(!Trout)Trout=O;
        else Trout->AddFriend("Anticounter");
    }else{
        cout << "Anticounter  : missing tree"<<endl;
    };
    // Orbital Info
    if(B && ORB) {
	B->SetBranchAddress("OrbitalInfo", GetPointerTo("OrbitalInfo"));
	cout << "OrbitalInfo  : set branch address OrbitalInfo"<<endl;
	if(!Trout)Trout=O;
	else Trout->AddFriend("OrbitalInfo");
    }else{
	cout << "OrbitalInfo  : missing tree"<<endl;
    };
    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;

    // Selection List
    if(L && SELLI==1) {
	cout<<">>> Found selection-list <<<"<<endl;
	L->SetBranchAddress("RunEntry",&irun);
	cout << "SelectionList: set branch address RunEntry"<<endl;
	L->SetBranchAddress("EventEntry",&irunentry);
	cout << "SelectionList: set branch address EventEntry"<<endl;
	sel_tree = L;
// 	if(!Trout)Trout=O;
// 	else Trout->AddFriend("SelectionList");
	cout << "----------------------------------------------------" <<endl;
    }else{
//	cout << "SelectionList  : missing tree"<<endl;
	if(L)L->Delete();
    };
    
//    cout<<endl<<" Number of entries: "<<Trout->GetEntries()<<endl<<endl;

    pam_tree = Trout;

    return Trout;
}



//--------------------------------------
//
//
//--------------------------------------
/**
 * Set branch addresses for Pamela friend trees
 */
void PamLevel2::SetBranchAddress(TTree *t){

    TRK2    = TRK2 & t->GetBranchStatus("TrkLevel2");
    TRK1    = TRK1 & t->GetBranchStatus("TrkLevel1");
    TRKh    = TRKh & t->GetBranchStatus("TrkHough");
    CAL2    = CAL2 & t->GetBranchStatus("CaloLevel2");
    CAL1    = CAL1 & t->GetBranchStatus("CaloLevel1");
    TOF    = TOF & t->GetBranchStatus("ToFLevel2");
    TRG    = TRG & t->GetBranchStatus("TrigLevel2");
    S4     = S4  & t->GetBranchStatus("S4Level2");
    ND     = ND  & t->GetBranchStatus("NDLevel2");
    AC     = AC  & t->GetBranchStatus("AcLevel2");
    ORB    = ORB & t->GetBranchStatus("OrbitalInfo");


    // Tracker
    if(TRK1) {
	t->SetBranchAddress("TrkLevel1", GetPointerTo("TrkLevel2"));
	cout << "Tracker      : set branch address TrkLevel1"<<endl;
    };
    if(TRK2) {
	t->SetBranchAddress("TrkLevel2", GetPointerTo("TrkLevel1"));
	cout << "Tracker      : set branch address TrkLevel2"<<endl;
    };
    if(TRKh) {
	t->SetBranchAddress("TrkHough", GetPointerTo("TrkHough"));
	cout << "Tracker      : set branch address TrkHough"<<endl;
    };
    
    // Calorimeter
    if(CAL1) {
	t->SetBranchAddress("CaloLevel1", GetPointerTo("CaloLevel1"));
	cout << "Calorimeter  : set branch address CaloLevel1"<<endl;
    };
    if(CAL2) {
	t->SetBranchAddress("CaloLevel2", GetPointerTo("CaloLevel2"));
	cout << "Calorimeter  : set branch address CaloLevel2"<<endl;
    };
    
    // ToF    
    if(TOF) {
	t->SetBranchAddress("ToFLevel2", GetPointerTo("ToFLevel2"));
	cout << "ToF          : set branch address ToFLevel2"<<endl;
    };
    // Trigger
    if(TRG) {
	t->SetBranchAddress("TrigLevel2", GetPointerTo("TrigLevel2"));
	cout << "Trigger      : set branch address TrigLevel2"<<endl;
    };
    // S4
    if(S4) {
	t->SetBranchAddress("S4Level2", GetPointerTo("S4Level2"));
	cout << "S4           : set branch address S4Level2"<<endl;
    };
    // Neutron Detector
    if(ND) {
	t->SetBranchAddress("NDLevel2", GetPointerTo("NDLevel2"));
	cout << "NeutronD     : set branch address NDLevel2"<<endl;
    };
    // Anticounters
    if(AC) {
	t->SetBranchAddress("AcLevel2", GetPointerTo("AcLevel2"));
	cout << "Anticounter  : set branch address AcLevel2"<<endl;
    };
    // OrbitalInfo
    if(ORB) {
	t->SetBranchAddress("OrbitalInfo", GetPointerTo("OrbitalInfo"));
	cout << "OrbitalInfo  : set branch address OrbitalInfo"<<endl;
    };
    // SelectionList
    if(SELLI==1) {
	t->SetBranchAddress("RunEntry", &irun);
	cout << "SelectionList: set branch address RunEntry"<<endl;
	t->SetBranchAddress("EventEntry", &irunentry);
	cout << "SelectionList: set branch address EventEntry"<<endl;
    };
    
}
/**
 * Set branch addresses for Pamela friend trees
 */
void PamLevel2::SetBranchAddress(TChain *t){

    TRK2    = TRK2 & t->GetBranchStatus("TrkLevel2");
    TRK1    = TRK1 & t->GetBranchStatus("TrkLevel1");
    TRKh    = TRKh & t->GetBranchStatus("TrkHough");
    CAL1    = CAL1 & t->GetBranchStatus("CaloLevel1");
    CAL2    = CAL2 & t->GetBranchStatus("CaloLevel2");
    TOF    = TOF & t->GetBranchStatus("ToFLevel2");
    TRG    = TRG & t->GetBranchStatus("TrigLevel2");
    S4     = S4  & t->GetBranchStatus("S4Level2");
    ND     = ND  & t->GetBranchStatus("NDLevel2");
    AC     = AC  & t->GetBranchStatus("AcLevel2");
    ORB    = ORB & t->GetBranchStatus("OrbitalInfo");

    // Tracker
     if(TRK2) {
	t->SetBranchAddress("TrkLevel2", GetPointerTo("TrkLevel2"));
	cout << "Tracker      : set branch address TrkLevel2"<<endl;
    };
    if(TRK1) {
	t->SetBranchAddress("TrkLevel1", GetPointerTo("TrkLevel1"));
	cout << "Tracker      : set branch address TrkLevel1"<<endl;
    };
     if(TRKh) {
 	t->SetBranchAddress("TrkHough", GetPointerTo("TrkHough"));
 	cout << "Tracker      : set branch address TrkHough"<<endl;
     };
    
    // Calorimeter
    if(CAL2) {
	t->SetBranchAddress("CaloLevel2", GetPointerTo("CaloLevel2"));
	cout << "Calorimeter  : set branch address CaloLevel2"<<endl;
    };   
    if(CAL1) {
	t->SetBranchAddress("CaloLevel1", GetPointerTo("CaloLevel1"));
	cout << "Calorimeter  : set branch address CaloLevel1"<<endl;
    };
    
    // ToF    
    if(TOF) {
	t->SetBranchAddress("ToFLevel2", GetPointerTo("ToFLevel2"));
	cout << "ToF          : set branch address ToFLevel2"<<endl;
    };
    // Trigger
    if(TRG) {
	t->SetBranchAddress("TrigLevel2", GetPointerTo("TrigLevel2"));
	cout << "Trigger      : set branch address TrigLevel2"<<endl;
    };
    // S4
    if(S4) {
	t->SetBranchAddress("S4Level2", GetPointerTo("S4Level2"));
	cout << "S4           : set branch address S4Level2"<<endl;
    };
    // Neutron Detector
    if(ND) {
	t->SetBranchAddress("NDLevel2", GetPointerTo("NDLevel2"));
	cout << "NeutronD     : set branch address NDLevel2"<<endl;
    };
    // Anticounters
    if(AC) {
	t->SetBranchAddress("AcLevel2", GetPointerTo("AcLevel2"));
	cout << "Anticounter  : set branch address AcLevel2"<<endl;
    };
    // OrbitalInfo
    if(ORB) {
	t->SetBranchAddress("OrbitalInfo", GetPointerTo("OrbitalInfo"));
	cout << "OrbitalInfo  : set branch address OrbitalInfo"<<endl;
    };
    // SelectionList
    if(SELLI==1) {
	t->SetBranchAddress("RunEntry", &irun);
	cout << "SelectionList: set branch address RunEntry"<<endl;
	t->SetBranchAddress("EventEntry", &irunentry);
	cout << "SelectionList: set branch address EventEntry"<<endl;
    };
    
}


//--------------------------------------
//
//
//--------------------------------------
/**
 * Get the Run tree chain from a list of files.
 * @param fl Pointer to a TList of TSystemFiles 
 * @return Pointer to a TChain
 */
TChain *PamLevel2::GetRunTree(TList *fl){
  //
  //
  //
  if ( run_tree ){
    printf("WARNING: TChain *PamLevel2::GetRunTree(TList *fl) -- run_tree already exists!\n ");
    return run_tree;
  };	
  //
  TChain *R = new TChain("Run");
    
    // loop over files and create chains	
    TIter next(fl);
    TSystemFile *questo = 0;
    while ( (questo = (TSystemFile*) next()) ) {
	TString name =  questo->GetName();
//		cout << "File: "<< name << endl;
	if( CheckLevel2File(name) ){
	    R->Add(name);
	};
    }

    if(R->GetNtrees()){
    
	R->SetBranchAddress("RunInfo", GetPointerTo("RunInfo"));
	cout << "Run         : set branch address RunInfo"<<endl;
	R->SetBranchAddress("SoftInfo", GetPointerTo("SoftInfo")); // Emiliano
	cout << "Software    : set branch address SoftInfo"<<endl; // Emiliano
    }else{
	delete R;
	R=0;
    }

    run_tree = R;

    return R;
    
}
//--------------------------------------
//
//
//--------------------------------------
/**
 * Get the Run tree  from a file.
 * @param f Pointer to a TFile 
 * @return Pointer to a TTree
 */
TTree *PamLevel2::GetRunTree(TFile *f){
  if ( run_tree ){
    printf("WARNING: TTree *PamLevel2::GetRunTree(TFile *f) -- run_tree already exists!\n ");
    return run_tree;
  };
	

    cout << "TTree *PamLevel2::GetRunTree(TFile *f) -- obsolte "<<endl;

    TTree *T = (TTree*)f->Get("Run");
    
    if(T){
	T->SetBranchAddress("RunInfo", GetPointerTo("RunInfo"));
	cout << "Run         : set branch address RunInfo"<<endl;
	T->SetBranchAddress("SoftInfo", GetPointerTo("SoftInfo")); // Emiliano
	cout << "Software    : set branch address SoftInfo"<<endl; // Emiliano
    }

    run_tree = (TChain*)T;

    return T;
    
}
/**
 * Update the runinfo informations (to be used to have Run infos event by event basis)
 * @param run Pointer to the chain/tree which contains run infos 
 * @return true if a new run has been read, false if it is still the same run
 */
Bool_t PamLevel2::UpdateRunInfo(TChain *run, Long64_t iev){
  //
  // check if we have already called once GetEntry, if not call it
  //
    cout << "Bool_t PamLevel2::UpdateRunInfo(TChain *run, Long64_t iev) --- NON FUNZIONA BENISSIMO.... "<<endl;
    if(!run){
	  cout << "Bool_t PamLevel2::UpdateRunInfo(TChain *run, ULong64_t iev) -- ERROR -- missing RunInfo tree "<<endl;
	  return(false);	
    }
    if ( run->GetEntries() <= 0 ) return(false);
  //
  
  //  Int_t oldrun = irun;
  Long64_t oldrun = irun;

  // --------------------------------------
  // if it is a full file (not preselected)
  // --------------------------------------
  if(SELLI==0){

      //
      // the absolute time is necessary to relate the event with the run
      //
      if( !GetOrbitalInfo() && !ISGP){
	  cout << "Bool_t PamLevel2::UpdateRunInfo(TChain *run, ULong64_t iev) -- ERROR -- missing OrbitalInfo "<<endl;
	  return(false);
      }

      ULong64_t abstime = 0;
      if( GetOrbitalInfo() )abstime=GetOrbitalInfo()->absTime;


      //
      // the first time the routine is called, set run search from the beginning
      //
      if ( irun < 0LL ){
	  irun = 0LL; 
	  run->GetEntry(irun);
 	  runfirstentry = 0LL;
 	  runlastentry += (Long64_t)(this->GetRunInfo()->NEVENTS);
 	  if ( (Long64_t)(this->GetRunInfo()->NEVENTS) > 0LL ) runlastentry -= 1LL;
	  
	  if( ISGP && run->GetEntries()!=1 ){
	      cout << "** WARNING ** simulated files are assumed to have 1 single run, not "<< run->GetEntries() << endl;
	      cout << "** WARNING ** run will not be updated"<<endl;
	  }

      };      
      //
      if(ISGP)abstime = GetRunInfo()->RUNHEADER_TIME; // BARBATRUCCO
      //
      if ( irun == run->GetEntries()-1LL && 
	   !(abstime >= GetRunInfo()->RUNHEADER_TIME && 
	     abstime <= GetRunInfo()->RUNTRAILER_TIME) 
	    ){
	 irun = -1LL;
	 runfirstentry = 0LL;
	 runlastentry = -1LL;
       };
      // modificato il controllo sull'aggiornamento del run, per evitare problemi
      // dovuti agli eventi annidati (NB! NEVENTS conta anche questi!!)
      //
      bool fromfirst = true;
      //
      while ( !(abstime >= GetRunInfo()->RUNHEADER_TIME && abstime <= GetRunInfo()->RUNTRAILER_TIME) && irun < run->GetEntries()-1LL ){
//      while ( iev > (runfirstentry+(ULong64_t)(this->GetRunInfo()->NEVENTS-1)) && irun < run->GetEntries() ){
	  irun++;
	  run->GetEntry(irun);
	  runfirstentry = runlastentry;
	  if ( (Long64_t)(this->GetRunInfo()->NEVENTS) > 0LL ) runfirstentry += 1LL;
	  runlastentry += (Long64_t)(this->GetRunInfo()->NEVENTS);
// 	  cout << " ))))) UPDATE RUN INFO (((((  @iev "<<iev<<" run "<<GetRunInfo()->ID<<" irun "<<irun<<endl;
// 	  cout << "runfirstentry "<<runfirstentry<<endl;
	  //	  printf(" iev %llu %u %llu \n",iev,this->GetRunInfo()->NEVENTS,(ULong64_t)(runfirstentry+(ULong64_t)(this->GetRunInfo()->NEVENTS)));
	  //	  printf(" abstime %u trailertime %u \n",abstime,GetRunInfo()->RUNTRAILER_TIME);
	  //	  printf(" IDRUN %u \n",GetRunInfo()->ID);
	  //
// 	  prevshift = 0;
	  //
	  if ( irun == (Long64_t)(run->GetEntries()-1LL) && fromfirst && !(abstime >= GetRunInfo()->RUNHEADER_TIME && abstime <= GetRunInfo()->RUNTRAILER_TIME)){
	      printf(" resetting irun  (it should NOT happen!!!)\n");
	      fromfirst = false;
	      irun = 0;
	      run->GetEntry(irun);
	      runfirstentry = 0ULL;
	      runlastentry += (Long64_t)(this->GetRunInfo()->NEVENTS);
	      if ( (Long64_t)(this->GetRunInfo()->NEVENTS) > 0LL ) runlastentry -= 1LL;
	  };
	  //
      };
      //
      if ( 
	   !(abstime >= GetRunInfo()->RUNHEADER_TIME && 
	     abstime <= GetRunInfo()->RUNTRAILER_TIME) 
	  ) {
	printf(" Something very wrong here: cannot find RUN containing absolute time %llu \n",abstime);
	  return false;
      }
      //
      if ( irun == oldrun || irun >= run->GetEntries() ) return(false);
      //
      //  printf(" iev %llu irun %i nevents %u 1st %llu last %llu \n",iev,irun,this->GetRunInfo()->NEVENTS,(ULong64_t)runfirstentry,(ULong64_t)runlastentry);
      //
      prevshift = 0;
      cout << " ))))) UPDATE RUN INFO (((((  @iev "<<iev<<" run "<<GetRunInfo()->ID<<" irun "<<irun<<endl;
//      cout << "runfirstentry "<<runfirstentry<<endl;
      return(true);    
  };
  // ----------------------------------------------------
  // if it is a preselected file (there is SelectionList)
  // NBNB - the event tree MUST be read first 
  // ----------------------------------------------------
  if(SELLI==1){      
      sel_tree->GetEntry(iev);
//      cout << irun << " "<< irunentry << endl;
      if(irun != oldrun){
	  run->GetEntry(irun);
	  cout << " ))))) UPDATE RUN INFO (((((  @iev "<<iev<<" run "<<GetRunInfo()->ID<<" irun "<<irun<<endl;
	  prevshift = 0;
	  return true;
      }
      return false;
  }

  return false;
  //
};

Bool_t PamLevel2::UpdateRunInfo(Long64_t iev){

    if(!run_tree){
	cout << " Bool_t PamLevel2::UpdateRunInfo(ULong64_t iev) -- ERROR -- run tree not loaded"<<endl;
	return false;
    }
    if ( run_tree->GetEntries() <= 0 ) {
	cout << " Bool_t PamLevel2::UpdateRunInfo(ULong64_t iev) -- ERROR -- run tree is empty"<<endl;
	return(false);
    }
  
    Int_t oldrun = irun; // store current run index

    // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    // if it is a full file (not preselected)
    // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    if(SELLI==0){

	ULong64_t abstime = 0;
	ULong64_t obt     = 0;
	if( GetOrbitalInfo() ){
	    abstime = GetOrbitalInfo()->absTime;
	    obt     = GetOrbitalInfo()->OBT;
	}
	// ---------------------------------------------------------------
	// the absolute time is necessary to relate the event with the run
	// ---------------------------------------------------------------
	if( !GetOrbitalInfo() && !ISGP ){
	    cout << "Bool_t PamLevel2::UpdateRunInfo(Long64_t iev) -- ERROR -- missing OrbitalInfo "<<endl;
	    return(false);
	}
      
	// -----------------------------------------------------------------------
	// the first time the routine is called, set run search from the beginning
	// -----------------------------------------------------------------------
	if ( irun < 0 ){
	    irun = 0LL;
	    irunentry = 0;
	    prevshift = 0;
	    run_tree->GetEntry(irun);

	    if( ISGP && run_tree->GetEntries()!=1 ){
		cout << "** WARNING ** simulated files are assumed to have 1 single run, not "<< run_tree->GetEntries() << endl;
		cout << "** WARNING ** run will not be updated"<<endl;
	    }
	};      

	// -----------------------------------------------------------------------
	// if it is simulation, assigh abstime by end (temporaneo!!!)
	// -----------------------------------------------------------------------
	if(ISGP){
	    abstime = GetRunInfo()->RUNHEADER_TIME; // BARBATRUCCO
	    obt     = GetRunInfo()->RUNHEADER_OBT; // BARBATRUCCO
	}
	//
	bool fromfirst = true; // first loop over runs

//	Bool_t hasfrag = false;
//	if( GetRunInfo()->ID_RUN_FRAG!=0 && GetRunInfo()->ID_RUN_FRAG != GetRunInfo()->ID)hasfrag=true;
//	ULong64_t fragid = GetRunInfo()->ID_RUN_FRAG;

	// ------------------------------------------------------
	// loop over runs to find the one that contains the event 
	// ------------------------------------------------------
	while ( 
	    (
	        (
		    !(abstime >= GetRunInfo()->RUNHEADER_TIME &&    // check on absolute time (s)
		      abstime <= GetRunInfo()->RUNTRAILER_TIME) && 
		    !(obt >= GetRunInfo()->RUNHEADER_OBT &&         // additional check on OBT (ms)
		      obt <= GetRunInfo()->RUNTRAILER_OBT)
		    ) 
  		|| GetRunInfo()->NEVENTS==0 
		|| !(irunentry < GetRunInfo()->NEVENTS-1-prevshift)
		)
		&& irun < run_tree->GetEntries() ){

	    irun++;
	    // ------------------------------------
	    // if the end of run tree is reached...
	    // ------------------------------------
	    if( irun == run_tree->GetEntries() ){
		if(!fromfirst){
		    // -----------------------------------------------------
		    // if it happened already once and the run was not found
		    // ---> exit with error
		    // -----------------------------------------------------
		    cout << "Bool_t PamLevel2::UpdateRunInfo(Long64_t iev) -- ERROR -- event entry #"<<iev<<" does not belong to any run (should not happen)" <<endl;
		    return false;
		}
		// -----------------------------------------
		// ...otherwise repeat search from beginning
		// -----------------------------------------
		cout << "Bool_t PamLevel2::UpdateRunInfo(Long64_t iev) -- WARNING -- reached end of run tree. searchin again from beginning " <<endl;
		fromfirst = false;
		irun = 0LL;
		runfirstentry = 0LL;
	    }
	    // -------------------------------------------------------------------
	    // save the index of the first entry of the run, relative to pam_tree,
	    // and read a new run
	    // -------------------------------------------------------------------
	    if(irun>0)runfirstentry += (GetRunInfo()->NEVENTS)-prevshift;
	    irunentry = 0;
	    prevshift = 0;
	    run_tree->GetEntry(irun);	   
	    if(GetRunInfo()->RUNHEADER_OBT>GetRunInfo()->RUNTRAILER_OBT ){
		cout << "Bool_t PamLevel2::UpdateRunInfo(Long64_t iev) -- WARNING -- irun "<<irun<<"  has RUNHEADER_OBT>=RUNTRAILER_OBT " <<endl;
		cout << "                                                            (NB!! in this case some events are assigned to a wrong run)"<<endl;
	    }
//  	    if(hasfrag &&  fragid != GetRunInfo()->ID){
//  		cout << "... where is the next fragment ??"<<endl;
// 	    }
	};


	// --------------------------------------
	// if there was no need to update the run
	// ---> exit with FALSE
	// --------------------------------------
	if ( irun == oldrun ) return(false);

	// --------------------------------------
	// ... otherwise 
	// ---> exit with TRUE
	// --------------------------------------
	cout << endl << " ))))) UPDATE RUN INFO (((((  @iev "<<iev<<" run "<<GetRunInfo()->ID<<" irun "<<irun<<endl;
	// ----------------------------------------------------
	// then check if the run has a fragment 
	// in this case we have to switch to the next fragment
	// when the end of the first fragment is reached
	// ----------------------------------------------------
	if( 
	    GetRunInfo()->ID_RUN_FRAG != 0 && 
//	    GetRunInfo()->ID_RUN_FRAG != GetRunInfo()->ID && 
	    true ){
	    cout << "* fragment *"<<endl;		
	} 
	
	return(true);    
    };
    // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    // if it is a preselected file (there is SelectionList)
    // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    if(SELLI==1){      
	sel_tree->GetEntry(iev);
	if(irun != oldrun){
	    run_tree->GetEntry(irun);
	    cout << endl << " ))))) UPDATE RUN INFO (((((  @iev "<<iev<<" run "<<GetRunInfo()->ID<<" irun "<<irun<<endl;
	    prevshift = 0;
	    return true;
	}
	return false;
    }

    return false;
    //
};
/**
 * Update the runinfo informations (to be used to have Run infos event by event basis)
 * @param run Pointer to the chain/tree which contains run infos 
 * @return true if a new run has been read, false if it is still the same run
 */
Bool_t PamLevel2::UpdateRunInfo(TTree *run, Long64_t iev){
  return(UpdateRunInfo((TChain*)run,iev));
};

//--------------------------------------
//
//
//--------------------------------------
/**
 * Set which trees shoul be analysed
 * @param detlist TString containing the sequence of trees required 
*/
void PamLevel2::SetWhichTrees(TString detlist){
	
    if(detlist.IsNull() || detlist.Contains("+ALL", TString::kIgnoreCase)){
	CAL0 = false;
	CAL1 = true;
	CAL2 = true;
	TRK2 = true;
	TRK1 = false;
	TRKh = false;
	TRK0 = false;
	TRG = true;
	TOF = true;
	TOF0 = false;
	S4  = true;
	ND  = true;
	AC  = true;
	ORB = true;
    }else if( detlist.Contains("-ALL", TString::kIgnoreCase) ){
	CAL0 = false;
	CAL1 = false;
	CAL2 = false;
	TRK2 = false;
	TRK1 = false;
	TRKh = false;
	TRK0 = false;
	TRG = false;
	TOF = false;
	TOF0 = false;
	S4  = false;
	ND  = false;
	AC  = false;
	ORB = false;
    };
    
//  -------------------------------------------------------------------------
    if( detlist.Contains("CAL1", TString::kIgnoreCase) ){
	if ( detlist.Contains("-CAL1", TString::kIgnoreCase) )CAL1=false;
	if ( detlist.Contains("+CAL1", TString::kIgnoreCase) )CAL1=true;
    };

    if( detlist.Contains("CAL0", TString::kIgnoreCase) ){
	if ( detlist.Contains("-CAL0", TString::kIgnoreCase) )CAL0=false;
	if ( detlist.Contains("+CAL0", TString::kIgnoreCase) )CAL0=true;
    };
	
    if( detlist.Contains("CAL2", TString::kIgnoreCase)){
	if ( detlist.Contains("-CAL2", TString::kIgnoreCase) )CAL2=false;
	if ( detlist.Contains("+CAL2", TString::kIgnoreCase) )CAL2=true;
    };
	
    if( detlist.Contains("+CAL", TString::kIgnoreCase) && !CAL1 && !CAL2 )CAL2=true;
    if( detlist.Contains("-CAL", TString::kIgnoreCase) && CAL1 && CAL2 ){
	CAL2=false;
	CAL1=false;
    }
//  -------------------------------------------------------------------------
    if( detlist.Contains("TRK0", TString::kIgnoreCase) ){
	if ( detlist.Contains("-TRK0", TString::kIgnoreCase) )TRK0=false;
	if ( detlist.Contains("+TRK0", TString::kIgnoreCase) )TRK0=true;
    };

    if( detlist.Contains("TRK1", TString::kIgnoreCase) ){
	if ( detlist.Contains("-TRK1", TString::kIgnoreCase) )TRK1=false;
	if ( detlist.Contains("+TRK1", TString::kIgnoreCase) )TRK1=true;
    };
	
    if( detlist.Contains("TRK2", TString::kIgnoreCase)){
	if ( detlist.Contains("-TRK2", TString::kIgnoreCase) )TRK2=false;
	if ( detlist.Contains("+TRK2", TString::kIgnoreCase) )TRK2=true;
    };

    if( detlist.Contains("TRKh", TString::kIgnoreCase)){
	if ( detlist.Contains("-TRKh", TString::kIgnoreCase) )TRKh=false;
	if ( detlist.Contains("+TRKh", TString::kIgnoreCase) )TRKh=true;
    };
	
    if( detlist.Contains("+TRK", TString::kIgnoreCase) && !TRK1 && !TRK2 && !TRKh )TRK2=true;
    if( detlist.Contains("-TRK", TString::kIgnoreCase) && TRK1 && TRK2 && TRKh){
	TRK2=false;
	TRK1=false;
	TRKh=false;
    }
//  -------------------------------------------------------------------------
    
    if( detlist.Contains("-TRG", TString::kIgnoreCase) )TRG = false;
    else if( detlist.Contains("+TRG", TString::kIgnoreCase) )TRG = true;
    
    if( detlist.Contains("-TOF", TString::kIgnoreCase) )TOF = false;
    else if( detlist.Contains("+TOF", TString::kIgnoreCase) )TOF = true;

    if( detlist.Contains("-TOF0", TString::kIgnoreCase) )TOF0 = false;
    else if( detlist.Contains("+TOF0", TString::kIgnoreCase) )TOF0 = true;
    
    if( detlist.Contains("-S4",  TString::kIgnoreCase) )S4  = false;
    else if( detlist.Contains("+S4",  TString::kIgnoreCase) )S4  = true;
    
    if( detlist.Contains("-ND",  TString::kIgnoreCase) )ND  = false;
    else if( detlist.Contains("+ND",  TString::kIgnoreCase) )ND  = true;
    
    if( detlist.Contains("-AC",  TString::kIgnoreCase) )AC  = false;
    else if( detlist.Contains("+AC",  TString::kIgnoreCase) )AC  = true;
    
    if( detlist.Contains("-ORB", TString::kIgnoreCase) )ORB = false;
    else if( detlist.Contains("+ORB", TString::kIgnoreCase) )ORB = true;

//     cout<< "Set detector list --> ";
//     if(TRK1)cout<<"TRK1 ";
//     if(TRK2)cout<<"TRK2 ";
//     if(TRKh)cout<<"TRKH ";
//     if(CAL1)cout<<"CAL1 ";
//     if(CAL2)cout<<"CAL2 ";
//     if(TOF)cout<<"TOF ";
//     if(TRG)cout<<"TRG ";
//     if(AC)cout<<"AC ";
//     if(ND)cout<<"ND ";
//     if(S4)cout<<"S4 ";
//     if(ORB)cout<<"ORB ";
//     cout << endl;
    
};


/**
 * Set tree/branch detector flags from the content of a tree
 */
void  PamLevel2::GetWhichTrees(TFile* f){
    


    cout << "void  PamLevel2::GetWhichTrees(TFile* f) --- WARNING!! --- ...potrebbe non funzionare "<<endl; 
    // -----------
    // reset flags
    // -----------
    CAL1   = false;     
    CAL2   = false;     
    TRK2   = false;     
    TRK1   = false;     
    TRKh   = false;     
    TRG    = false;     
    TOF    = false;     
    S4     = false;    
    ND     = false;    
    AC     = false;     
    ORB    = false;    
    
    RUN    = false;
        
    cout << "Checking file: "<<f->GetName()<<endl;
    if( !f || f->IsZombie() ){
	cout << "File: "<< f->GetName() <<" Non valid root file"<< endl; 
	return;
    }

    TList *lk = f->GetListOfKeys();
    TIter next(lk);
    TKey *key =0;

    Int_t nev = 0;

    while( (key = (TKey*)next()) ){
	
	if( !strcmp(key->GetName(),"Run"        ) )RUN = true;

	//=========================================================	
       	if( !strcmp(key->GetName(),"Trigger"    ) ){ 
	    TRG = true;
	    Int_t nevt = ((TTree*)f->Get("Trigger"))->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<" Trigger tree has "<<nevt<<" events instead of "<<nev<< endl; 
		TRG = false;
	    }else nev=nevt;
	}
	//=========================================================	
	if( !strcmp(key->GetName(),"ToF"        ) ){
	    TOF = true;
	    Int_t nevt = ((TTree*)f->Get("ToF"))->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<"     ToF tree has "<<nevt<<" events instead of "<<nev<< endl; 
		TOF = false;
	    }else nev=nevt;
	}
	//=========================================================  
	if( !strcmp(key->GetName(),"S4"         ) ){
	    S4 = true;
	    Int_t nevt = ((TTree*)f->Get("S4"))->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<"      S4 tree has "<<nevt<<" events instead of "<<nev<< endl; 
		S4 = false;
	    }else nev=nevt;
	}
	//=========================================================  

	if( !strcmp(key->GetName(),"NeutronD"   ) ){
	    ND = true;
	    Int_t nevt = ((TTree*)f->Get("NeutronD"))->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<"NeutronD tree has "<<nevt<<" events instead of "<<nev<< endl; 
		ND =false;
	    }else nev=nevt;
	}	
        //=========================================================  
	if( !strcmp(key->GetName(),"Anticounter") ){
	    AC = true;
	    Int_t nevt = ((TTree*)f->Get("Anticounter"))->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<" Anticounter tree has "<<nevt<<" events instead of "<<nev<< endl; 
		AC =false;
	    }else nev=nevt;
	}
	//=========================================================  
	if( !strcmp(key->GetName(),"OrbitalInfo") ){
	    ORB = true;
	    Int_t nevt = ((TTree*)f->Get("OrbitalInfo"))->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<" OrbitalInfo tree has "<<nevt<<" events instead of "<<nev<< endl; 
		ORB = false;
	    }else nev=nevt;
	}
	//=========================================================  
	if( !strcmp(key->GetName(),"Tracker"    ) ){
	    TTree *T = (TTree*)f->Get("Tracker");
	    for(Int_t i=0; i<T->GetListOfBranches()->GetEntries(); i++){
		TString name = T->GetListOfBranches()->At(i)->GetName();
		if( !name.CompareTo("TrkLevel1") )TRK1=true;
		if( !name.CompareTo("TrkLevel2") )TRK2=true;
		if( !name.CompareTo("TrkHough") )TRKh=true;
	    };	    
	    Int_t nevt = T->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<" Tracker tree has "<<nevt<<" events instead of "<<nev<< endl; 
		TRK1 = false;
		TRK2 = false;
		TRKh = false;
	    }else nev=nevt;
	    T->Delete();
	};
	//=========================================================  
	if( !strcmp(key->GetName(),"Calorimeter"    ) ){
	    TTree *T = (TTree*)f->Get("Calorimeter");
	    for(Int_t i=0; i<T->GetListOfBranches()->GetEntries(); i++){
		TString name = T->GetListOfBranches()->At(i)->GetName();
		if( !name.CompareTo("CaloLevel1") )CAL1=true;
		if( !name.CompareTo("CaloLevel2") )CAL2=true;
	    };	   
	    Int_t nevt = T->GetEntries();
	    if( nev && nevt!=nev){
		cout << "File: "<< f->GetName() <<"  Calorimeter tree has "<<nevt<<" events instead of "<<nev<< endl; 
		CAL1 = false;
		CAL2 = false;
	    }else nev=nevt;
	    T->Delete();
	};	

    };
    
    delete lk;

//     cout<< "Get detector list from input file --> ";
//     if(TRK1)cout<<"TRK1 ";
//     if(TRK2)cout<<"TRK2 ";
//     if(TRKh)cout<<"TRKH ";
//     if(CAL1)cout<<"CAL1 ";
//     if(CAL2)cout<<"CAL2 ";
//     if(TOF)cout<<"TOF ";
//     if(TRG)cout<<"TRG ";
//     if(AC)cout<<"AC ";
//     if(ND)cout<<"ND ";
//     if(S4)cout<<"S4 ";
//     if(ORB)cout<<"ORB ";
//     cout << endl;
       
    return ;
	
};


//--------------------------------------
//
//
//--------------------------------------
/**
 * Check if a file contains selected Pamela Level2 trees.
 * @param name File name
 * @return true if the file is ok.
 */
Bool_t  PamLevel2::CheckLevel2File(TString name){
	
    Bool_t CAL1__ok   = false;     
    Bool_t CAL2__ok   = false;     
    Bool_t TRK2__ok   = false;     
    Bool_t TRK1__ok   = false;     
    Bool_t TRKh__ok   = false;     
    Bool_t TRG__ok    = false;     
    Bool_t TOF__ok    = false;     
    Bool_t S4__ok     = false;    
    Bool_t ND__ok     = false;    
    Bool_t AC__ok     = false;     
    Bool_t ORB__ok    = false;    
    
    Bool_t RUN__ok    = false;
    
    Bool_t SELLI__ok  = false;

    cout << "Checking file: "<<name<<endl;
    TFile *f = new TFile(name.Data());
    if( !f || f->IsZombie() ){
	cout << "File: "<< f->GetName() <<" discarded ---- Non valid root file"<< endl; return false;
    }
//    cout << "Get list of keys: "<<f<<endl;
    TList *lk = f->GetListOfKeys();
//    lk->Print();
    TIter next(lk);
    TKey *key =0;

    Int_t nev = 0;

    while( (key = (TKey*)next()) ){
	
// 	cout << key->GetName() << endl;
// 	cout << key->GetName() << ""<<key->GetClassName()<<endl;
//		cout << " Get tree: " << f->Get(key->GetName())<<endl;
// 	nev_previous = nev;
// 	cout << " n.entries  "<< nev <<endl;
// 	if( key->GetClassName()=="TTree" && nev_previous && nev != nev_previous ){
// 	    nev = ((TTree*)f->Get(key->GetName()))->GetEntries();
// 	    cout << "File: "<< f->GetName() <<" discarded ---- "<< key->GetName() << " tree: n.entries does not match "<<nev<<" "<<nev_previous<< endl;
// 	    return false;
// 	};


	if( !strcmp(key->GetName(),"Run"        ) )RUN__ok = true;	

	//=========================================================	
       	if( !strcmp(key->GetName(),"SelectionList"    ) ){ 
	    SELLI__ok = true;
	    if(SELLI==1){
		Int_t nevt = ((TTree*)f->Get("SelectionList"))->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- SelectionList tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	}

	//=========================================================	
       	if( !strcmp(key->GetName(),"Trigger"    ) ){ 
	    TRG__ok = true;
	    if(TRG){
		Int_t nevt = ((TTree*)f->Get("Trigger"))->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- Trigger tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	}
	//=========================================================	
	if( !strcmp(key->GetName(),"ToF"        ) ){
	    TOF__ok = true;
	    if(TOF){
		Int_t nevt = ((TTree*)f->Get("ToF"))->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- ToF tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	}
	//=========================================================  
	if( !strcmp(key->GetName(),"S4"         ) ){
	    S4__ok = true;
	    if(S4){
		Int_t nevt = ((TTree*)f->Get("S4"))->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- S4 tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	}
	//=========================================================  

	if( !strcmp(key->GetName(),"NeutronD"   ) ){
	    ND__ok = true;
	    if(ND){
		Int_t nevt = ((TTree*)f->Get("NeutronD"))->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- NeutronD tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	}	
        //=========================================================  
	if( !strcmp(key->GetName(),"Anticounter") ){
	    AC__ok = true;
	    if(AC){
		Int_t nevt = ((TTree*)f->Get("Anticounter"))->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- Anticounter tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	}
	//=========================================================  
	if( !strcmp(key->GetName(),"OrbitalInfo") ){
	    ORB__ok = true;
	    if(ORB){
		Int_t nevt = ((TTree*)f->Get("OrbitalInfo"))->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- OrbitalInfo tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	}
	//=========================================================  
	if( !strcmp(key->GetName(),"Tracker"    ) ){
	    TTree *T = (TTree*)f->Get("Tracker");
	    if(TRK1||TRK2||TRKh){
		Int_t nevt = T->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- Tracker tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	    for(Int_t i=0; i<T->GetListOfBranches()->GetEntries(); i++){
		TString name = T->GetListOfBranches()->At(i)->GetName();
		if( !name.CompareTo("TrkLevel1") )TRK1__ok=true;
		if( !name.CompareTo("TrkLevel2") )TRK2__ok=true;
		if( !name.CompareTo("TrkHough") )TRKh__ok=true;
	    };	    
	    T->Delete();
	};
	//=========================================================  
	if( !strcmp(key->GetName(),"Calorimeter"    ) ){
	    TTree *T = (TTree*)f->Get("Calorimeter");
	    if(CAL1||CAL2){
		Int_t nevt = T->GetEntries();
		if( nev && nevt!=nev){
		    cout << "File: "<< f->GetName() <<" discarded ---- Calorimeter tree has "<<nevt<<" events instead of "<<nev<< endl; 
		    return false;
		}
		nev=nevt;
	    }
	    for(Int_t i=0; i<T->GetListOfBranches()->GetEntries(); i++){
		TString name = T->GetListOfBranches()->At(i)->GetName();
		if( !name.CompareTo("CaloLevel1") )CAL1__ok=true;
		if( !name.CompareTo("CaloLevel2") )CAL2__ok=true;
	    };	   
	    T->Delete();
	};	

    };

    if( SELLI==-1 )SELLI = (Int_t)SELLI__ok;
    if( SELLI==0 && SELLI__ok ){
	cout << "File: "<< f->GetName() <<" discarded ---- found SelectionList (it is not a full-event file)" << endl; 
	return false;   
    }
    if( SELLI==1 && !SELLI__ok ){
	cout << "File: "<< f->GetName() <<" discarded ---- SelectionList missing" << endl; 
	return false;
    }
    
//    cout << "SELLI "<<SELLI<<endl;

//     cout<< "CheckLevel2File(TString): detector list --> ";
//     if(TRK1__ok)cout<<"TRK1 ";
//     if(TRK2__ok)cout<<"TRK2 ";
//     if(TRKh__ok)cout<<"TRKH ";
//     if(CAL1__ok)cout<<"CAL1 ";
//     if(CAL2__ok)cout<<"CAL2 ";
//     if(TOF__ok)cout<<"TOF ";
//     if(TRG__ok)cout<<"TRG ";
//     if(AC__ok)cout<<"AC ";
//     if(ND__ok)cout<<"ND ";
//     if(S4__ok)cout<<"S4 ";
//     if(ORB__ok)cout<<"ORB ";
//     cout << endl;


    if(TRK2 && TRK1__ok)TRK1=1; 
// ----------------------------------------------------------------------------
// NOTA
// se c'e` il level1, lo devo necessarimente leggere.
// infatti (non ho capito perche`) i cluster vengono letti e allocati in memoria 
// comunque, ma non vengono disallocati da PamLevel2::Clear()
// ----------------------------------------------------------------------------


    if(!RUN__ok) {
	cout << "File: "<< f->GetName() <<" *WARNING* ---- Missing RunInfo tree"<< endl; 
//	return false;
    };

    if(CAL1 && !CAL1__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing CaloLevel1 branch"<< endl; 
	return false;
    };
    if(CAL2 && !CAL2__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing CaloLevel2 branch"<< endl; 
	return false;
    };
    if(TRK2 && !TRK2__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing TrkLevel2 branch"<< endl; 
	return false;
    };
    if(TRK1 && !TRK1__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing TrkLevel1 branch"<< endl; 
	return false;
    };
    if(TRKh && !TRKh__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing TrkHough branch"<< endl;
	return false;
    };
    if(ORB && !ORB__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing ORB tree"<< endl; 
	return false;
    };
    if(AC && !AC__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing AC tree"<< endl; 
	return false;
    };
    if(S4 && !S4__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing S4 tree"<< endl; 
	return false;
    };
    if(TOF && !TOF__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing ToF tree"<< endl; 
	return false;
    };

    if(ND && !ND__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing ND tree"<< endl;
	return false;
    };
    if(TRG && !TRG__ok){
	cout << "File: "<< f->GetName() <<" discarded ---- Missing Trigger tree"<< endl; 
	return false;
    };
    

//    lk->Delete();
//    delete lk;
    f->Close();

//     cout<< "CheckLevel2File(TString): detector list --> ";
//     if(TRK1)cout<<"TRK1 ";
//     if(TRK2)cout<<"TRK2 ";
//     if(TRKh)cout<<"TRKH ";
//     if(CAL1)cout<<"CAL1 ";
//     if(CAL2)cout<<"CAL2 ";
//     if(TOF)cout<<"TOF ";
//     if(TRG)cout<<"TRG ";
//     if(AC)cout<<"AC ";
//     if(ND)cout<<"ND ";
//     if(S4)cout<<"S4 ";
//     if(ORB)cout<<"ORB ";
//     cout << endl;
       
    return true;
	
};


/**
 * Create clone-trees
 */
void PamLevel2::CreateCloneTrees0( TChain *fChain, TFile *ofile ){

    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;
    cout << "Create clones of PAMELA trees "<<endl;
    
    Int_t i=0;
    pam_tree_clone[i] = fChain->GetTree()->CloneTree(0);
    TString name =  pam_tree_clone[i]->GetName();
    name.Append("_clone");
//    pam_tree_clone[i]->SetName(name.Data());
    cout << pam_tree_clone[i]->GetName() <<endl;
    i++;

    TList *li = fChain->GetListOfFriends();
    TIter next(li);
    TFriendElement* T_friend=0;
    ofile->cd();
    while( (T_friend = (TFriendElement*)next()) ){
// 	cout<<T_friend->IsA()->GetName()<<" "<<T_friend->GetName()<<hex << T_friend->GetTree() << dec<<endl;
//	cout<<T_friend->GetTree()->GetName()<< endl;
	pam_tree_clone[i] = T_friend->GetTree()->CloneTree(0);
	pam_tree_clone[i]->SetAutoSave(1000000);
	name =  pam_tree_clone[i]->GetName();
	name.Append("_clone");
//	pam_tree_clone[i]->SetName(name.Data());
	cout << pam_tree_clone[i]->GetName() << endl;
	i++;
    }
    
    delete li;

    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;

}

/**
 * Create clone-trees
 */
void PamLevel2::CreateCloneTrees(TFile *ofile){


//  if the pointer is null, create a default file

    if(!ofile){
	cout << "void PamLevel2::CreateCloneTrees(TFile*) -- WARNING -- Creating file: clone-tree.root "<<endl;
	ofile = new TFile("clone-tree.root","recreate");
    }

    ofile->cd();

    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;
    cout << "Create new PAMELA trees "<<endl;
        

    run_tree_clone = new TTree("Run","PAMELA Level2 data from the GL_RUN table ");
    run_tree_clone->Branch("RunInfo","GL_RUN",GetPointerTo("RunInfo"));
    cout << "Run          : branch RunInfo"<<endl;
    run_tree_clone->Branch("SoftInfo","SoftInfo",GetPointerTo("SoftInfo"));
    cout << "Run          : branch SoftInfo"<<endl;
    // ------------------
    // replicate run tree
    // ------------------
    cout << "----------------------------------------------------"<<endl;
    cout << "irun\t | RUN\t NEVENTS\t absolute time"<<endl;
    for (Int_t i=0; i<run_tree->GetEntries(); i++){
	run_tree->GetEntry(i);
	cout << i<< "\t | "<<GetRunInfo()->ID<<"\t "<<GetRunInfo()->NEVENTS<< "\t "<<GetRunInfo()->RUNHEADER_TIME<<" <---> "<<GetRunInfo()->RUNTRAILER_TIME<<endl;
	run_tree_clone->Fill();
    }
    cout << "----------------------------------------------------"<<endl;


    sel_tree_clone = new TTree("SelectionList","List of selected events ");
    sel_tree_clone->Branch("RunEntry",&irun,"runentry/L");
    sel_tree_clone->Branch("EventEntry",&irunentry,"eventry/L");
    

    Int_t i=0;
    if(TRK1||TRK2||TRKh){
	pam_tree_clone[i] = new TTree("Tracker","PAMELA tracker level2 data ");	
	if(TRK1) {
	    pam_tree_clone[i]->Branch("TrkLevel1","TrkLevel1", GetPointerTo("TrkLevel1"));
	    pam_tree_clone[i]->BranchRef();
	    cout << "Tracker      : branch TrkLevel1"<<endl;
//	    cout << "CreateCloneTrees " << GetTrkLevel1()<<endl;
	};
	if(TRK2) {
	    pam_tree_clone[i]->Branch("TrkLevel2", "TrkLevel2",GetPointerTo("TrkLevel2"));
	    cout << "Tracker      : branch TrkLevel2"<<endl;	    
	};
	if(TRKh) {
	    pam_tree_clone[i]->Branch("TrkHough","TrkHough", GetPointerTo("TrkHough"));
	    cout << "Tracker      : branch TrkHough"<<endl;
	};
	i++;
    }

    // Calorimeter
    if(CAL1||CAL2){
	pam_tree_clone[i] = new TTree("Calorimeter","PAMELA calorimeter level2 data ");	
	if(CAL1) {
	    pam_tree_clone[i]->Branch("CaloLevel1", "CaloLevel1", GetPointerTo("CaloLevel1"));
	    cout << "Calorimeter  : branch CaloLevel1"<<endl;
	};
	if(CAL2) {
	    pam_tree_clone[i]->Branch("CaloLevel2","CaloLevel2", GetPointerTo("CaloLevel2"));
	    cout << "Calorimeter  : branch CaloLevel2"<<endl;
	};
	i++;
    }    

    // ToF    
    if(TOF) {
	pam_tree_clone[i] = new TTree("ToF","PAMELA ToF level2 data ");	
	pam_tree_clone[i]->Branch("ToFLevel2","ToFLevel2", GetPointerTo("ToFLevel2"));
	cout << "ToF          : branch ToFLevel2"<<endl;
	i++;
    };
    // Trigger
    if(TRG) {
	pam_tree_clone[i] = new TTree("Trigger","PAMELA trigger level2 data ");	
	pam_tree_clone[i]->Branch("TrigLevel2","TrigLevel2", GetPointerTo("TrigLevel2"));
	cout << "Trigger      : branch TrigLevel2"<<endl;
	i++;
    };
    // S4
    if(S4) {
	pam_tree_clone[i] = new TTree("S4","PAMELA S4 level2 data ");	
	pam_tree_clone[i]->Branch("S4Level2","S4Level2", GetPointerTo("S4Level2"));
	cout << "S4           : branch S4Level2"<<endl;
	i++;
    };
    // Neutron Detector
    if(ND) {
	pam_tree_clone[i] = new TTree("NeutronD","PAMELA neutron detector level2 data ");	
	pam_tree_clone[i]->Branch("NDLevel2","NDLevel2", GetPointerTo("NDLevel2"));
	cout << "NeutronD     : branch NDLevel2"<<endl;
	i++;
    };
    // Anticounters
    if(AC) {
	pam_tree_clone[i] = new TTree("Anticounter","PAMELA anticounter detector level2 data ");	
	pam_tree_clone[i]->Branch("AcLevel2","AcLevel2", GetPointerTo("AcLevel2"));
	cout << "Anticounter  : branch AcLevel2"<<endl;
	i++;
    };
    // OrbitalInfo
    if(ORB) {
	pam_tree_clone[i] = new TTree("OrbitalInfo","PAMELA oribital info  ");	
	pam_tree_clone[i]->Branch("OrbitalInfo","OrbitalInfo", GetPointerTo("OrbitalInfo"));
	cout << "OrbitalInfo  : branch OrbitalInfo"<<endl;
	i++;
    };
    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;

}


/**
 * Fill tree (created with CreateCloneTrees)
 *
 */
//void PamLevel2::FillNewPamTree(TTree *T){
void PamLevel2::FillCloneTrees(){
    
//    cout << "PamLevel2::FillCloneTrees()" << irunentry << endl;

    for(Int_t i=0; i<NCLONES; i++){
	if(pam_tree_clone[i])pam_tree_clone[i]->Fill();
    }
    if(sel_tree_clone)sel_tree_clone->Fill();    

}


TTree* PamLevel2::GetCloneTree(TString name){

    for(Int_t i=0; i<NCLONES; i++){	
	if(pam_tree_clone[i]){
	    TString na = pam_tree_clone[i]->GetName();
	    if(!name.CompareTo(na))return pam_tree_clone[i];
	};
    }
    if(run_tree_clone){
	TString na = run_tree_clone->GetName();
	if(!name.CompareTo(na))return run_tree_clone;
    }
    if(sel_tree_clone){
	TString na = sel_tree_clone->GetName();
	if(!name.CompareTo(na))return sel_tree_clone;
    }
    return NULL;

}
void PamLevel2::WriteCloneTrees(){
    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;
    cout << "Write clones of PAMELA trees "<<endl;
    cout << run_tree_clone->GetName()<<endl;    
    run_tree_clone->Write();
    cout << sel_tree_clone->GetName()<<endl;    
    sel_tree_clone->Write();
    for(Int_t i=0; i<NCLONES; i++){	
	if(pam_tree_clone[i]){
	    cout << pam_tree_clone[i]->GetName()<<endl;
	    pam_tree_clone[i]->Write();
	};
    }
    cout << "+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+" <<endl;

}

/**
 * Method to get level2-trees entry, the corresponding run entry and (if required) the level0 entry.
 */
//Int_t PamLevel2::GetEntry(Int_t iee){
Int_t PamLevel2::GetEntry(Long64_t iee){
    
//     cout << "-------------------------------------"<<endl;
//     cout << "Int_t PamLevel2::GetEntry("<<iee<<")"<<endl;

    if(!pam_tree){
	cout << " Int_t PamLevel2::GetEntry(Int_t) -- ERROR -- level2 trees not loaded"<<endl;
	return 0;
    }


    //
    // This is a sort of bug: if you don't have the run tree you don't want to exit here you want to have loaded the event anyway...
    //
    //    if(!run_tree ){
    //	cout << " Int_t PamLevel2::GetEntry(Int_t) -- ERROR -- run tree not loeaded"<<endl;
    //	return 0;
    //    }

    Long64_t ii=0;
    //-------------------------------
    ii = iee;
    if( !pam_tree->GetEntry(ii) ){	
	cout << " Int_t PamLevel2::GetEntry(Int_t) -- ERROR -- error reading pam tree"<<endl;
	return 0;
    }
    //
    // ... that's way I put it here. Notice that nothing change in the code (is backward compatible) since in any case you return with 0. 
    // in theory one would like to return 1 if run is not loaded but now I don't have the will to add that 2 lines of code and it is not
    // a problem if you don't check the return code of getentry.
    //
    if(!run_tree ){
	if ( TRK0 || CAL0 || TOF0 || RUN ) { //forse cosi` va bene per tornare 1?
	    cout << " Int_t PamLevel2::GetEntry(Int_t) -- ERROR -- run tree not loaded"<<endl;
	    return 0;
	}  else {
	    return 1; //cosi` se non c'e` run esce qua...
	}
    }

    //-------------------------------
    ii = iee;
//    Bool_t UPDATED = UpdateRunInfo(run_tree,ii);
//    Bool_t UPDATED = UpdateRunInfo(ii);
    UpdateRunInfo(ii);
    if(SELLI==0)irunentry = iee-runfirstentry;
//    if(UPDATED && run_tree_clone)run_tree_clone->Fill();
   
//    cout << "PamLevel2::GetEntry("<<iee<<") "<<irun<<" "<<runfirstentry<<" "<<irunentry<<endl;

//     cout << "irunentry     "<<irunentry << endl;
//     cout << "runfirstentry "<<runfirstentry << endl;
//     cout << "nevents       "<<GetRunInfo()->NEVENTS<< endl;

//     if( TRK0 || CAL0 || TOF0 ){
// 	if( !GetYodaEntry( ) ){
// 	    cout << " Int_t PamLevel2::GetEntry(Int_t) -- ERROR -- error reading level0 tree"<<endl;
// 	    return 0;
// 	}
//     }


    return 1;

}

TrkLevel0    *PamLevel2::GetTrkLevel0(){
    if( !TRK0 )return NULL;
    if( !GetYodaEntry( ) ){
	cout << " Int_t PamLevel2::GetTrkLevel0() -- ERROR -- error reading level0 tree"<<endl;
	return 0;
    }
    return trk0_obj;
};
CaloLevel0    *PamLevel2::GetCaloLevel0(){
    if( !CAL0 )return NULL;
    if( !GetYodaEntry( ) ){
	cout << " Int_t PamLevel2::GetCaloLevel0() -- ERROR -- error reading level0 tree"<<endl;
	return 0;
    }
    return calo0_obj;
};


/**
 * Method to retrieve the level0 tree (YODA tree) that contains the current event.
 * Given the run ID (...), if needed it query the DB and load the proper file.
 * @return Pointer to the tree
 */


TTree* PamLevel2::GetYodaTree( ){
    
//    cout << "TTree* PamLevel2::GetYodaTree( )"<<endl;
    //===================================
    // check if iroot has changed
    //===================================
    if( irun<0 ){
	cout << "PamLevel2::GetYodaTree() -- ERROR -- irun = "<<irun<<endl;
//	cout << "In order to use this method you have to first load the RunInfo tree "<<endl;
	return NULL;
    }    
    Int_t irootnew = run_obj->ID_ROOT_L0;
//      cout << "iroot    "<<iroot<<endl;
//      cout << "irootnew "<<irootnew<<endl;

    //===================================
    // load the level0 file
    // (if not already loaded)
    //===================================
    if( iroot != irootnew || !l0_tree){
	iroot = irootnew;
	//===================================
	// open the DB connection
	// (if not already opened)
	//===================================
	if(!dbc || (dbc && !dbc->IsConnected())){
	    cout<<"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"<<endl; 
	    cout<<"Connecting to DB"<<endl;
	    cout<<"HOST "<<host<<endl;
	    cout<<"USER "<<user<<endl;
	    cout<<"PSW  "<<psw<<endl;
	    dbc = TSQLServer::Connect(host.Data(),user.Data(),psw.Data());
	    if( !dbc )return NULL;
	    if( !dbc->IsConnected() )return NULL;    
//	    cout<<"...done"<<endl;
	    cout<<"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"<<endl; 
	}else{
//	    cout<<"DB already connected"<<endl;
	}
	GL_ROOT glroot = GL_ROOT();
	if( glroot.Query_GL_ROOT(iroot,dbc) ){
	    cout << "TTree* PamLevel2::GetYodaTree( ) -- ERROR -- level0 file iroot = "<<iroot<< " does not exists"<<endl;
	    return NULL;
	};
	TString filename = glroot.PATH + glroot.NAME;
	if(l0_file){ 
	    l0_file->Close();
	    l0_file->Delete();
	}
	cout << "Opening LEVEL0 file: "<< filename << endl;
	FileStat_t t;
	if( gSystem->GetPathInfo(filename.Data(),t) ){
	    cout << " PamLevel2::GetYodaTree() -- ERROR opening file "<<endl;
	    return NULL;
	}
	l0_file = new TFile(filename);
	if( !l0_file )return NULL;
	l0_tree = (TTree*)l0_file->Get("Physics");
	if(!h0_obj)h0_obj = new EventHeader();
	l0_tree->SetBranchAddress("Header" ,&h0_obj);
	prevshift = 0;
	//---------------------------------------------------
	// TRACKER:
	if(TRK0){
	    if(!trk0_obj){
		trk0_obj = new TrkLevel0();
		trk0_obj->Set();
	    };
	    l0_tree->SetBranchAddress("Tracker" ,trk0_obj->GetPointerToTrackerEvent());
	}
	//--------------------------------------------------
	// CALORIMETER:
	if(CAL0){
	    if(!calo0_obj){
		calo0_obj = new CaloLevel0();
		calo0_obj->Set();
	    };
	    l0_tree->SetBranchAddress("Calorimeter" ,calo0_obj->GetPointerToCalorimeterEvent());
	    //	    cout << "PamLevel2::GetYodaTree() --- level0 calorimeter not implemented "<<endl;
	}
	//---------------------------------------------------
	// TOF:
	if(TOF0){
	    cout << "PamLevel2::GetYodaTree() --- level0 TOF not implemented "<<endl;
	}

    };

    if(!dbc || (dbc && !dbc->IsConnected())){
	cout << " TTree* PamLevel2::GetYodaTree( ) -- no DB connected... hai fatto qualche cazzata "<<endl;
    }

    if ( TRK0 ){
	TrkParams::Load(6);
	if( !TrkParams::IsLoaded(6) ){
	    cout << " TTree* PamLevel2::GetYodaTree( ) -- WARNING -- VK-mask not loaded"<<endl;
	};
	TrkParams::SetCalib(run_obj,dbc);
	TrkParams::LoadCalib( );
	if( !TrkParams::CalibIsLoaded() ){
	    cout << " TTree* PamLevel2::GetYodaTree( ) -- WARNING -- Calibration not loaded"<<endl;
	};
    }

//    cout << l0_tree << endl;
    
    return l0_tree;

}

/**
 * Method to retrieve the level0 tree (YODA tree) that contains the current event.
 */
Int_t PamLevel2::GetYodaEntry(){

//    cout << "Int_t PamLevel2::GetYodaEntry()"<<endl;
    if(!GetYodaTree())return 0;
    
    // patch
    if( irunentry < 0){
//	cout << "Int_t PamLevel2::GetYodaEntry() -- ATTENZIONE -- irunentry negativo?!?! "<<(Int_t)irunentry<<endl;
	irunentry=0LL;
    }
    //  ---------------------------------
    //  if file is NOT a preselected file
    //  ---------------------------------
    Long64_t quellagiusta = irunentry + (Long64_t)(run_obj->EV_FROM);
//     cout << " irunentry "<<irunentry<<endl;
//     cout << " EV_FROM "<<run_obj->EV_FROM<<endl;
//     cout << " quellagiusta = irunentry + EV_FROM "<< quellagiusta << endl;

//     cout << " irun "<< irun << " irunentry "<< irunentry<<" run_obj->EV_FROM "<<run_obj->EV_FROM <<" quella giusta "<<quellagiusta << endl;
//     cout << " iroot "<<iroot<<" run_obj->ID_ROOT_L0 "<<run_obj->ID_ROOT_L0<<endl;
//     cout << " time "<< abstime << endl;
//     cout << " trk_calib_used "<<run_obj->TRK_CALIB_USED<< endl;
    
    ULong64_t obt     = 0;
    ULong64_t pktn    = 0;
    if( GetOrbitalInfo() ){ 
	obt     = GetOrbitalInfo()->OBT;
	pktn    = GetOrbitalInfo()->pkt_num;
    }

    if( !GetOrbitalInfo() && !ISGP){
	cout << "Int_t PamLevel2::GetYodaEntry() -- ERROR -- missing OrbitalInfo "<<endl;
	return 0;
    }
    if( obt==0 && pktn==0 && !ISGP ){
	cout << "Int_t PamLevel2::GetYodaEntry() -- ERROR -- level2 event corrupted ?? "<<endl;	
	return 0;
    }

    // ---------------------------------------------------------------------
    // ATTENTION!!!
    // If data are not pre-processed with cleaner, the level0 tree may contain
    // spurious nested physics packets.
    // The GL_RUN variables EV_FROM, EV_TO and NEVENTS counts also these entries
    // while level2 tree DOES NOT!!
    // This means that "quellagiusta" in these cases is not correct.
    // In order to retrieve the correct level0 event, I implemented a check
    // of the OBT and pkt-number. In case of mismatch, the level0 entry number
    // is shift forward until when the packets match.
    // ---------------------------------------------------------------------
    Int_t answer = 0;
    Int_t shift =0;
    //    printf(" siamo qui %i %i \n",shift,prevshift);
    Int_t maxshift = 10;
    do{
	if(shift>0){	 
	    cout << " PKTNUM  L2 --- "<< pktn << " --- L0 --- "<< GetEventHeader()->GetPscuHeader()->GetCounter()<<endl;
	    cout << "         RUN: ID "<< GetRunInfo()->ID << " ID_ROOT_L0 "<<run_obj->ID_ROOT_L0<<" ID_RUN_FRAG "<<GetRunInfo()->ID_RUN_FRAG << " EV_FROM "<<GetRunInfo()->EV_FROM  <<endl;
	    cout << "         L2 <--> L0 mismatch ( irun "<<irun<<" irunentry "<<irunentry<<" shift "<<shift<<" prevshift "<<prevshift<<" )"<<endl;
	}
	answer = l0_tree->GetEntry(quellagiusta+(Long64_t)shift+(Long64_t)prevshift);
	shift++;
	if( !GetEventHeader() ){
	    cout << "Int_t PamLevel2::GetYodaEntry() -- ERROR -- missing EventHeader "<<endl;
	    return 0;
	}

	if(ISGP){
	    obt = (UInt_t)(GetEventHeader()->GetPscuHeader()->GetOrbitalTime()); //BARBATRUCCO
	    pktn = (UInt_t)(GetEventHeader()->GetPscuHeader()->GetCounter()); //BARBATRUCCO
	}

//	cout << "PKTNUM "<<shift<<" ==  L2 --- "<< pktn << " --- L0 --- "<< GetEventHeader()->GetPscuHeader()->GetCounter()<<endl;
//	cout << " L2 --- "<< obt << " --- L0 --- "<< GetEventHeader()->GetPscuHeader()->GetOrbitalTime()<<endl;
// 	if( (quellagiusta+shift) == l0_tree->GetEntries()+1 )cout << ">>> end of level0 tree <<<"<<endl; 
// 	cout << " obt "<< obt << endl;
// 	cout << " GetEventHeader()->GetPscuHeader()->GetOrbitalTime() "<< GetEventHeader()->GetPscuHeader()->GetOrbitalTime() << endl;
// 	cout << " pktn "<< pktn << endl;
// 	cout << " GetEventHeader()->GetPscuHeader()->GetCounter() "<< GetEventHeader()->GetPscuHeader()->GetCounter() << endl;
//	printf(" IDRUN %u \n",GetRunInfo()->ID);
//
	if ( prevshift != 0 && (quellagiusta+(Long64_t)shift) == GetYodaTree()->GetEntries() ){
	  prevshift = 0;
	  shift = -1;
	};

    }while( ( obt != (UInt_t)(GetEventHeader()->GetPscuHeader()->GetOrbitalTime()) || 
	      pktn != (UInt_t)(GetEventHeader()->GetPscuHeader()->GetCounter())       ) 
	    && (quellagiusta+(Long64_t)shift) < GetYodaTree()->GetEntries() && shift < maxshift);

    if ( (quellagiusta+(Long64_t)shift+(Long64_t)prevshift) > GetYodaTree()->GetEntries() || shift == maxshift ) {
	cout << " Big trouble here, no such event in Level0 data! " <<endl;
	return 0;
    }
//    cout << "LA ENTRY GIUSTA E`: "<<quellagiusta<<" (spero...)"<<endl;
//    return GetYodaTree()->GetEntry(quellagiusta);
    if ( shift > 1 ) prevshift += (shift-1);
    
    return answer;
    
}
/**
 * \Brief Set DB connection
 */
Bool_t PamLevel2::SetDBConnection(){

    cout<<"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"<<endl; 
    cout<<"Connecting to DB"<<endl;
    cout<<"HOST "<<host<<endl;
    cout<<"USER "<<user<<endl;
    cout<<"PSW  "<<psw<<endl;
    dbc = TSQLServer::Connect(host.Data(),user.Data(),psw.Data());
    if( !dbc )return false;
    if( !dbc->IsConnected() )return false;    
    cout<<"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"<<endl; 
    return true;

}
